from abc import ABC, abstractmethod
from typing import List

from canopy.knowledge_base.models import KBDocChunk
from canopy.models.data_models import Document
from canopy.utils.config import ConfigurableMixin


class Chunker(ABC, ConfigurableMixin):

    """
    Base class for chunkers. Chunkers take a document (id, text, ...)
    and return a list of KBDocChunks  (id, text, document_id, ...)
    Chunker is an abstract class that must be subclassed to be used,
    also, it extends ConfigurableMixin which means that every subclass of
    Chunker could be referenced by a name and configured in a config file.
    """

    def chunk_documents(self, documents: List[Document]) -> List[KBDocChunk]:
        """
        chunk_documents takes a list of documents and returns a list of KBDocChunks
        this method is just a wrapper around chunk_single_document that can be
        used to chunk a list of documents.

        Args:
            documents: list of documents

        Returns:
            chunks: list of chunks of type KBDocChunks
        """
        chunks: List[KBDocChunk] = []
        for doc in documents:
            chunks.extend(self.chunk_single_document(doc))
        return chunks

    async def achunk_documents(self, documents: List[Document]) -> List[KBDocChunk]:
        chunks: List[KBDocChunk] = []
        for doc in documents:
            chunks.extend(await self.achunk_single_document(doc))
        return chunks

    @abstractmethod
    def chunk_single_document(self, document: Document) -> List[KBDocChunk]:
        """
        chunk_single_document takes a document and returns a
        list of KBDocChunks, this is the main method
        that must be implemented by every subclass of Chunker

        Args:
            document: list of documents

        Returns:
            chunks: list of chunks KBDocChunks
        """
        pass

    @abstractmethod
    async def achunk_single_document(self, document: Document) -> List[KBDocChunk]:
        raise NotImplementedError()



from typing import List
from functools import cached_property
from pinecone_text.dense.base_dense_ecoder import BaseDenseEncoder

from .base import RecordEncoder
from canopy.knowledge_base.models import KBQuery, KBEncodedDocChunk, KBDocChunk
from canopy.models.data_models import Query


class DenseRecordEncoder(RecordEncoder):
    """
    DenseRecordEncoder is a subclass of RecordEncoder that generates dense vector representation of documents chunks and textual queries.
    The dense represntation generated by the `DenseRecordEncoder` is a list of floats in a given dimension.
    DenseRecordEncoder wraps a BaseDenseEncoder from the `pinecone-text` library to encode the text itself.
    for more information about the BaseDenseEncoder see: https://github.com/pinecone-io/pinecone-text
    """  # noqa: E501

    def __init__(self,
                 dense_encoder: BaseDenseEncoder,
                 **kwargs):
        """
        Initialize the encoder.

        Args:
            dense_encoder: A BaseDenseEncoder to encode the text.
            **kwargs: Additional arguments to pass to the RecordEncoder.
        """  # noqa: E501
        super().__init__(**kwargs)
        self._dense_encoder = dense_encoder

    def _encode_documents_batch(self,
                                documents: List[KBDocChunk]
                                ) -> List[KBEncodedDocChunk]:
        """
        Encode a batch of documents, takes a list of KBDocChunk and returns a list of KBEncodedDocChunk.

        Args:
            documents: A list of KBDocChunk to encode.
        Returns:
            encoded chunks: A list of KBEncodedDocChunk, with the `values` field populated by the generated embeddings vector.
        """  # noqa: E501
        dense_values = self._dense_encoder.encode_documents([d.text for d in documents])
        return [KBEncodedDocChunk(**d.dict(), values=v) for d, v in
                zip(documents, dense_values)]

    def _encode_queries_batch(self, queries: List[Query]) -> List[KBQuery]:
        """
        Encode a batch of queries, takes a list of Query and returns a list of KBQuery.
        Args:
            queries: A list of Query to encode.
        Returns:
            encoded queries: A list of KBQuery, with the `values` field populated by the generated embeddings vector.
        """  # noqa: E501
        dense_values = self._dense_encoder.encode_queries([q.text for q in queries])
        return [KBQuery(**q.dict(), values=v) for q, v in zip(queries, dense_values)]

    @cached_property
    def dimension(self) -> int:
        """
        The dimension is the length of the vector generated by the `DenseRecordEncoder`
        Canopy will run a single word through the encoder to get the dimension, this will also validate that the encoder
        is working properly.

        Returns:
            dimension(int): the dimension of the encoder
        """  # noqa: E501
        dummy_doc = KBDocChunk(text="hello", id="dummy_doc", document_id="dummy_doc")
        return len(self.encode_documents([dummy_doc])[0].values)

    async def _aencode_documents_batch(self,
                                       documents: List[KBDocChunk]
                                       ) -> List[KBEncodedDocChunk]:
        raise NotImplementedError

    async def _aencode_queries_batch(self, queries: List[Query]) -> List[KBQuery]:
        raise NotImplementedError



from typing import Optional, Sequence, Iterable

from pydantic import BaseModel, Field, validator

from canopy.models.data_models import MessageBase


class _Choice(BaseModel):
    index: int
    message: MessageBase
    finish_reason: Optional[str] = None


class _StreamChoice(BaseModel):
    index: int
    delta: dict
    finish_reason: Optional[str] = None


class TokenCounts(BaseModel):
    prompt_tokens: int
    completion_tokens: int
    total_tokens: Optional[int] = None

    @validator("total_tokens", always=True)
    def calc_total_tokens(cls, v, values, **kwargs):
        return values["prompt_tokens"] + values["completion_tokens"]


class ChatResponse(BaseModel):
    id: str = Field(description="Canopy session Id.")
    object: str
    created: int
    model: str
    choices: Sequence[_Choice]
    usage: TokenCounts
    debug_info: dict = Field(default_factory=dict, exclude=True)


class StreamingChatChunk(BaseModel):
    id: str
    object: str
    created: int
    model: str
    choices: Sequence[_StreamChoice]


class StreamingChatResponse(BaseModel):
    chunks: Iterable[StreamingChatChunk]
    debug_info: dict = Field(default_factory=dict, exclude=True)



import tiktoken
from typing import List
from .base import BaseTokenizer
from ..models.data_models import Messages


class OpenAITokenizer(BaseTokenizer):
    """
    Tokenizer for OpenAI models, based on the tiktoken library.

    Usage:
    Initialize the singleton tokenizer with the OpenAITokenizer class:
    >>> from canopy.tokenizer import Tokenizer
    >>> Tokenizer.initialize(tokenizer_class=OpenAITokenizer, model_name="gpt-3.5-turbo")

    You can then use the tokenizer instance from anywhere in the code:
    >>> from canopy.tokenizer import Tokenizer
    >>> tokenizer = Tokenizer()
    >>> tokenizer.tokenize("Hello world!")
    ['Hello', ' world', '!']
    """  # noqa: E501

    MESSAGE_TOKENS_OVERHEAD = 3
    FIXED_PREFIX_TOKENS = 3

    def __init__(self, model_name: str = "gpt-3.5-turbo"):
        """
        Initialize the tokenizer.

        Args:
            model_name: The name of the model to use. Defaults to "gpt-3.5-turbo".
                        You can find the list of available models here: https://github.com/openai/tiktoken/blob/39f29cecdb6fc38d9a3434e5dd15e4de58cf3c80/tiktoken/model.py#L19C1-L19C18
                        As you can see, both gpt-3.5 and gpt-4 are using the same cl100k_base tokenizer.
        """  # noqa: E501
        self._encoder = tiktoken.encoding_for_model(model_name)

    def tokenize(self, text: str) -> List[str]:
        """
        Tokenize a text using tiktoken.

        Args:
            text: The text to tokenize.

        Returns:
            The list of tokens.
        """
        return [self._encoder.decode([encoded_token])
                for encoded_token in self._encode(text)]

    def detokenize(self, tokens: List[str]) -> str:
        """
        Detokenize a list of tokens that were previously tokenized using this tokenizer.

        Args:
            tokens: The list of tokens to detokenize.

        Returns:
            The detokenized text as a string.
        """
        if not isinstance(tokens, List):
            raise TypeError(f"detokenize expect List[str], got f{type(tokens)}")
        return "".join(tokens)

    def token_count(self, text: str) -> int:
        """
        Count the number of tokens in a text.

        Args:
            text: The text to count the tokens of.

        Returns:
            The number of tokens in the text.
        """
        return len(self._encode(text))

    def _encode(self, text):
        return self._encoder.encode(text, disallowed_special=())

    def messages_token_count(self, messages: Messages) -> int:
        """
        Count the number of tokens in a list of messages as expected to be counted by OpenAI models.
        Account for the overhead of the messages structure.
        Taken from: https://github.com/openai/openai-cookbook/.../How_to_format_inputs_to_ChatGPT_models.ipynb

        Args:
            messages: The list of messages to count the tokens of.

        Returns:
            The number of tokens in the messages, as expected to be counted by OpenAI models.
        """  # noqa: E501
        num_tokens = 0
        for message in messages:
            num_tokens += self.MESSAGE_TOKENS_OVERHEAD
            for key, value in message.dict().items():
                num_tokens += self.token_count(value)
        num_tokens += self.FIXED_PREFIX_TOKENS
        return num_tokens



# Provided under standard MIT license by Harrison Chase LangChain 2022-10-17
# https://github.com/hwchase17/langchain

from __future__ import annotations

import logging
import re
from abc import ABC, abstractmethod
from enum import Enum
from typing import (
    Any,
    Callable,
    Iterable,
    List,
    Optional,
)

logger = logging.getLogger(__name__)


def _split_text_with_regex(
    text: str, separator: str, keep_separator: bool
) -> List[str]:
    # Now that we have the separator, split the text
    if separator:
        if keep_separator:
            # The parentheses in the pattern keep the delimiters in the result.
            _splits = re.split(f"({separator})", text)
            splits = [_splits[i] + _splits[i + 1] for i in range(1, len(_splits), 2)]
            if len(_splits) % 2 == 0:
                splits += _splits[-1:]
            splits = [_splits[0]] + splits
        else:
            splits = re.split(separator, text)
    else:
        splits = list(text)
    return [s for s in splits if s != ""]


class TextSplitter(ABC):
    """Interface for splitting text into chunks."""

    def __init__(
        self,
        chunk_size: int = 4000,
        chunk_overlap: int = 200,
        length_function: Callable[[str], int] = len,
        keep_separator: bool = False,
        add_start_index: bool = False,
        strip_whitespace: bool = True,
    ) -> None:
        """Create a new TextSplitter.

        Args:
            chunk_size: Maximum size of chunks to return
            chunk_overlap: Overlap in characters between chunks
            length_function: Function that measures the length of given chunks
            keep_separator: Whether to keep the separator in the chunks
            add_start_index: If `True`, includes chunk's start index in metadata
            strip_whitespace: If `True`, strips whitespace from the start and end of
                              every document
        """
        if chunk_overlap > chunk_size:
            raise ValueError(
                f"Got a larger chunk overlap ({chunk_overlap}) than chunk size "
                f"({chunk_size}), should be smaller."
            )
        self._chunk_size = chunk_size
        self._chunk_overlap = chunk_overlap
        self._length_function = length_function
        self._keep_separator = keep_separator
        self._add_start_index = add_start_index
        self._strip_whitespace = strip_whitespace

    @abstractmethod
    def split_text(self, text: str) -> List[str]:
        """Split text into multiple components."""

    def _join_docs(self, docs: List[str], separator: str) -> Optional[str]:
        text = separator.join(docs)
        if self._strip_whitespace:
            text = text.strip()
        if text == "":
            return None
        else:
            return text

    def _merge_splits(self, splits: Iterable[str], separator: str) -> List[str]:
        # We now want to combine these smaller pieces into medium size
        # chunks to send to the LLM.
        separator_len = self._length_function(separator)

        docs = []
        current_doc: List[str] = []
        total = 0
        for d in splits:
            _len = self._length_function(d)
            if (
                total + _len + (separator_len if len(current_doc) > 0 else 0)
                > self._chunk_size
            ):
                if total > self._chunk_size:
                    logger.warning(
                        f"Created a chunk of size {total}, "
                        f"which is longer than the specified {self._chunk_size}"
                    )
                if len(current_doc) > 0:
                    doc = self._join_docs(current_doc, separator)
                    if doc is not None:
                        docs.append(doc)
                    # Keep on popping if:
                    # - we have a larger chunk than in the chunk overlap
                    # - or if we still have any chunks and the length is long
                    while total > self._chunk_overlap or (
                        total + _len + (separator_len if len(current_doc) > 0 else 0)
                        > self._chunk_size
                        and total > 0
                    ):
                        total -= self._length_function(current_doc[0]) + (
                            separator_len if len(current_doc) > 1 else 0
                        )
                        current_doc = current_doc[1:]
            current_doc.append(d)
            total += _len + (separator_len if len(current_doc) > 1 else 0)
        doc = self._join_docs(current_doc, separator)
        if doc is not None:
            docs.append(doc)
        return docs


class Language(str, Enum):
    """Enum of the programming languages."""

    CPP = "cpp"
    GO = "go"
    JAVA = "java"
    JS = "js"
    PHP = "php"
    PROTO = "proto"
    PYTHON = "python"
    RST = "rst"
    RUBY = "ruby"
    RUST = "rust"
    SCALA = "scala"
    SWIFT = "swift"
    MARKDOWN = "markdown"
    LATEX = "latex"
    HTML = "html"
    SOL = "sol"
    CSHARP = "csharp"


class RecursiveCharacterTextSplitter(TextSplitter):
    """Splitting text by recursively look at characters.

    Recursively tries to split by different characters to find one
    that works.
    """

    def __init__(
        self,
        separators: Optional[List[str]] = None,
        keep_separator: bool = True,
        is_separator_regex: bool = False,
        **kwargs: Any,
    ) -> None:
        super().__init__(keep_separator=keep_separator, **kwargs)
        self._separators = separators or ["\n\n", "\n", " ", ""]
        self._is_separator_regex = is_separator_regex

    def _split_text(self, text: str, separators: List[str]) -> List[str]:
        """Split incoming text and return chunks."""
        final_chunks = []
        # Get appropriate separator to use
        separator = separators[-1]
        new_separators = []
        for i, _s in enumerate(separators):
            _separator = _s if self._is_separator_regex else re.escape(_s)
            if _s == "":
                separator = _s
                break
            if re.search(_separator, text):
                separator = _s
                new_separators = separators[i + 1:]
                break

        _separator = separator if self._is_separator_regex else re.escape(separator)
        splits = _split_text_with_regex(text, _separator, self._keep_separator)

        # Now go merging things, recursively splitting longer texts.
        _good_splits = []
        _separator = "" if self._keep_separator else separator
        for s in splits:
            if self._length_function(s) < self._chunk_size:
                _good_splits.append(s)
            else:
                if _good_splits:
                    merged_text = self._merge_splits(_good_splits, _separator)
                    final_chunks.extend(merged_text)
                    _good_splits = []
                if not new_separators:
                    final_chunks.append(s)
                else:
                    other_info = self._split_text(s, new_separators)
                    final_chunks.extend(other_info)
        if _good_splits:
            merged_text = self._merge_splits(_good_splits, _separator)
            final_chunks.extend(merged_text)
        return final_chunks

    def split_text(self, text: str) -> List[str]:
        return self._split_text(text, self._separators)

    @classmethod
    def from_language(
        cls, language: Language, **kwargs: Any
    ) -> RecursiveCharacterTextSplitter:
        separators = cls.get_separators_for_language(language)
        return cls(separators=separators, is_separator_regex=True, **kwargs)

    @staticmethod
    def get_separators_for_language(language: Language) -> List[str]:
        if language == Language.CPP:
            return [
                # Split along class definitions
                "\nclass ",
                # Split along function definitions
                "\nvoid ",
                "\nint ",
                "\nfloat ",
                "\ndouble ",
                # Split along control flow statements
                "\nif ",
                "\nfor ",
                "\nwhile ",
                "\nswitch ",
                "\ncase ",
                # Split by the normal type of lines
                "\n\n",
                "\n",
                " ",
                "",
            ]
        elif language == Language.GO:
            return [
                # Split along function definitions
                "\nfunc ",
                "\nvar ",
                "\nconst ",
                "\ntype ",
                # Split along control flow statements
                "\nif ",
                "\nfor ",
                "\nswitch ",
                "\ncase ",
                # Split by the normal type of lines
                "\n\n",
                "\n",
                " ",
                "",
            ]
        elif language == Language.JAVA:
            return [
                # Split along class definitions
                "\nclass ",
                # Split along method definitions
                "\npublic ",
                "\nprotected ",
                "\nprivate ",
                "\nstatic ",
                # Split along control flow statements
                "\nif ",
                "\nfor ",
                "\nwhile ",
                "\nswitch ",
                "\ncase ",
                # Split by the normal type of lines
                "\n\n",
                "\n",
                " ",
                "",
            ]
        elif language == Language.JS:
            return [
                # Split along function definitions
                "\nfunction ",
                "\nconst ",
                "\nlet ",
                "\nvar ",
                "\nclass ",
                # Split along control flow statements
                "\nif ",
                "\nfor ",
                "\nwhile ",
                "\nswitch ",
                "\ncase ",
                "\ndefault ",
                # Split by the normal type of lines
                "\n\n",
                "\n",
                " ",
                "",
            ]
        elif language == Language.PHP:
            return [
                # Split along function definitions
                "\nfunction ",
                # Split along class definitions
                "\nclass ",
                # Split along control flow statements
                "\nif ",
                "\nforeach ",
                "\nwhile ",
                "\ndo ",
                "\nswitch ",
                "\ncase ",
                # Split by the normal type of lines
                "\n\n",
                "\n",
                " ",
                "",
            ]
        elif language == Language.PROTO:
            return [
                # Split along message definitions
                "\nmessage ",
                # Split along service definitions
                "\nservice ",
                # Split along enum definitions
                "\nenum ",
                # Split along option definitions
                "\noption ",
                # Split along import statements
                "\nimport ",
                # Split along syntax declarations
                "\nsyntax ",
                # Split by the normal type of lines
                "\n\n",
                "\n",
                " ",
                "",
            ]
        elif language == Language.PYTHON:
            return [
                # First, try to split along class definitions
                "\nclass ",
                "\ndef ",
                "\n\tdef ",
                # Now split by the normal type of lines
                "\n\n",
                "\n",
                " ",
                "",
            ]
        elif language == Language.RST:
            return [
                # Split along section titles
                "\n=+\n",
                "\n-+\n",
                "\n\\*+\n",
                # Split along directive markers
                "\n\n.. *\n\n",
                # Split by the normal type of lines
                "\n\n",
                "\n",
                " ",
                "",
            ]
        elif language == Language.RUBY:
            return [
                # Split along method definitions
                "\ndef ",
                "\nclass ",
                # Split along control flow statements
                "\nif ",
                "\nunless ",
                "\nwhile ",
                "\nfor ",
                "\ndo ",
                "\nbegin ",
                "\nrescue ",
                # Split by the normal type of lines
                "\n\n",
                "\n",
                " ",
                "",
            ]
        elif language == Language.RUST:
            return [
                # Split along function definitions
                "\nfn ",
                "\nconst ",
                "\nlet ",
                # Split along control flow statements
                "\nif ",
                "\nwhile ",
                "\nfor ",
                "\nloop ",
                "\nmatch ",
                "\nconst ",
                # Split by the normal type of lines
                "\n\n",
                "\n",
                " ",
                "",
            ]
        elif language == Language.SCALA:
            return [
                # Split along class definitions
                "\nclass ",
                "\nobject ",
                # Split along method definitions
                "\ndef ",
                "\nval ",
                "\nvar ",
                # Split along control flow statements
                "\nif ",
                "\nfor ",
                "\nwhile ",
                "\nmatch ",
                "\ncase ",
                # Split by the normal type of lines
                "\n\n",
                "\n",
                " ",
                "",
            ]
        elif language == Language.SWIFT:
            return [
                # Split along function definitions
                "\nfunc ",
                # Split along class definitions
                "\nclass ",
                "\nstruct ",
                "\nenum ",
                # Split along control flow statements
                "\nif ",
                "\nfor ",
                "\nwhile ",
                "\ndo ",
                "\nswitch ",
                "\ncase ",
                # Split by the normal type of lines
                "\n\n",
                "\n",
                " ",
                "",
            ]
        elif language == Language.MARKDOWN:
            return [
                "\n## ",
                "\n### ",
                "\n#### ",
                "\n##### ",
                "\n###### ",
                # Note the alternative syntax for headings (below) is not handled here
                # Heading level 2
                # ---------------
                # End of code block
                "```\n",
                # add end of table to langchain implementation
                '|\n\n'
                # Horizontal lines
                "\n\n***\n\n",
                "\n\n---\n\n",
                "\n\n___\n\n",
                # Note that this splitter doesn't handle horizontal lines defined
                # by *three or more* of ***, ---, or ___, but this is not handled
                "\n\n",
                "\n",
                " ",
                "",
            ]
        elif language == Language.LATEX:
            return [
                # First, try to split along Latex sections
                "\n\\\\chapter{",
                "\n\\\\section{",
                "\n\\\\subsection{",
                "\n\\\\subsubsection{",
                # Now split by environments
                "\n\\\\begin{enumerate}",
                "\n\\\\begin{itemize}",
                "\n\\\\begin{description}",
                "\n\\\\begin{list}",
                "\n\\\\begin{quote}",
                "\n\\\\begin{quotation}",
                "\n\\\\begin{verse}",
                "\n\\\\begin{verbatim}",
                # Now split by math environments
                "\n\\\begin{align}",
                "$$",
                "$",
                # Now split by the normal type of lines
                " ",
                "",
            ]
        elif language == Language.HTML:
            return [
                # First, try to split along HTML tags
                "<body",
                "<div",
                "<p",
                "<br",
                "<li",
                "<h1",
                "<h2",
                "<h3",
                "<h4",
                "<h5",
                "<h6",
                "<span",
                "<table",
                "<tr",
                "<td",
                "<th",
                "<ul",
                "<ol",
                "<header",
                "<footer",
                "<nav",
                # Head
                "<head",
                "<style",
                "<script",
                "<meta",
                "<title",
                "",
            ]
        elif language == Language.CSHARP:
            return [
                "\ninterface ",
                "\nenum ",
                "\nimplements ",
                "\ndelegate ",
                "\nevent ",
                # Split along class definitions
                "\nclass ",
                "\nabstract ",
                # Split along method definitions
                "\npublic ",
                "\nprotected ",
                "\nprivate ",
                "\nstatic ",
                "\nreturn ",
                # Split along control flow statements
                "\nif ",
                "\ncontinue ",
                "\nfor ",
                "\nforeach ",
                "\nwhile ",
                "\nswitch ",
                "\nbreak ",
                "\ncase ",
                "\nelse ",
                # Split by exceptions
                "\ntry ",
                "\nthrow ",
                "\nfinally ",
                "\ncatch ",
                # Split by the normal type of lines
                "\n\n",
                "\n",
                " ",
                "",
            ]
        elif language == Language.SOL:
            return [
                # Split along compiler information definitions
                "\npragma ",
                "\nusing ",
                # Split along contract definitions
                "\ncontract ",
                "\ninterface ",
                "\nlibrary ",
                # Split along method definitions
                "\nconstructor ",
                "\ntype ",
                "\nfunction ",
                "\nevent ",
                "\nmodifier ",
                "\nerror ",
                "\nstruct ",
                "\nenum ",
                # Split along control flow statements
                "\nif ",
                "\nfor ",
                "\nwhile ",
                "\ndo while ",
                "\nassembly ",
                # Split by the normal type of lines
                "\n\n",
                "\n",
                " ",
                "",
            ]

        else:
            raise ValueError(
                f"Language {language} is not supported! "
                f"Please choose from {list(Language)}"
            )



import os
from typing import List
from pinecone_text.dense.openai_encoder import OpenAIEncoder
from canopy.knowledge_base.models import KBDocChunk, KBEncodedDocChunk, KBQuery
from canopy.knowledge_base.record_encoder.dense import DenseRecordEncoder
from canopy.models.data_models import Query

ANYSCALE_BASE_URL = "https://api.endpoints.anyscale.com/v1"


class AnyscaleRecordEncoder(DenseRecordEncoder):
    """
    AnyscaleRecordEncoder is a type of DenseRecordEncoder that uses the OpenAI `embeddings` API.
    The implementation uses the `OpenAIEncoder` class from the `pinecone-text` library.
    For more information about see: https://github.com/pinecone-io/pinecone-text

    """  # noqa: E501
    """
    Initialize the AnyscaleRecordEncoder

    Args:
        api_key: The Anyscale Endpoint API Key
        base_url: The Base URL for Anyscale Endpoint
        model_name: The name of the Anyscale embeddings model to use for encoding. See https://docs.endpoints.anyscale.com/category/supported-models
        batch_size: The number of documents or queries to encode at once.
                    Defaults to 400.
        **kwargs: Additional arguments to pass to the underlying `pinecone-text. OpenAIEncoder`.
    """  # noqa: E501
    def __init__(self,
                 *,
                 api_key: str = "",
                 base_url: str = ANYSCALE_BASE_URL,
                 model_name: str = "thenlper/gte-large",
                 batch_size: int = 400,
                 **kwargs):

        ae_api_key = api_key or os.environ.get("ANYSCALE_API_KEY")
        if not ae_api_key:
            raise ValueError(
                "Anyscale API key is required to use Anyscale. "
                "Please provide it as an argument "
                "or set the ANYSCALE_API_KEY environment variable."
            )
        ae_base_url = base_url
        encoder = OpenAIEncoder(model_name,
                                base_url=ae_base_url, api_key=ae_api_key,
                                **kwargs)
        super().__init__(dense_encoder=encoder, batch_size=batch_size)

    def encode_documents(self, documents: List[KBDocChunk]) -> List[KBEncodedDocChunk]:
        """
        Encode a list of documents, takes a list of KBDocChunk and returns a list of KBEncodedDocChunk.

        Args:
            documents: A list of KBDocChunk to encode.

        Returns:
            encoded chunks: A list of KBEncodedDocChunk, with the `values` field populated by the generated embeddings vector.
        """  # noqa: E501
        return super().encode_documents(documents)

    async def _aencode_documents_batch(self,
                                       documents: List[KBDocChunk]
                                       ) -> List[KBEncodedDocChunk]:
        raise NotImplementedError

    async def _aencode_queries_batch(self, queries: List[Query]) -> List[KBQuery]:
        raise NotImplementedError



import click
from click import ClickException

from canopy_cli.data_loader.data_loader import format_multiline


class CLIError(ClickException):
    def format_message(self) -> str:
        return click.style(format_multiline(self.message), fg='red')


class ConfigError(RuntimeError):
    pass



from typing import Tuple

from canopy.chat_engine.history_pruner.base import HistoryPruner
from canopy.models.data_models import Messages


class RaisingHistoryPruner(HistoryPruner):

    def build(self,
              history: Messages,
              max_tokens: int) -> Tuple[Messages, int]:
        token_count = self._tokenizer.messages_token_count(history)
        if token_count > max_tokens:
            raise ValueError(f"The history require {token_count} tokens, "
                             f"which exceeds the calculated limit for history "
                             f"of {max_tokens} tokens.")
        return history, token_count

    async def abuild(self,
                     full_history: Messages,
                     max_tokens: int) -> Tuple[Messages, int]:
        raise NotImplementedError



import importlib.metadata
import warnings
import logging
import os
from typing import List

# Taken from https://stackoverflow.com/a/67097076
__version__ = importlib.metadata.version("canopy-sdk")


IGNORED_WARNINGS: List[str] = [
]

IGNORED_WARNING_IN_MODULES = [
    "transformers",
]

for warning in IGNORED_WARNINGS:
    warnings.filterwarnings("ignore", message=warning)
for module in IGNORED_WARNING_IN_MODULES:
    warnings.filterwarnings("ignore", module=module)
    logging.getLogger(module).setLevel(logging.ERROR)

# Apparently, `transformers` has its own logging system, and needs to be silenced separately # noqa: E501
os.environ["TRANSFORMERS_VERBOSITY"] = "error"



from abc import ABC, abstractmethod
from typing import Union, Iterable, Optional, List

from canopy.llm.models import Function
from canopy.models.api_models import ChatResponse, StreamingChatChunk
from canopy.models.data_models import Messages, Query
from canopy.utils.config import ConfigurableMixin


class BaseLLM(ABC, ConfigurableMixin):
    def __init__(self,
                 model_name: str):
        self.model_name = model_name

    @abstractmethod
    def chat_completion(self,
                        messages: Messages,
                        *,
                        stream: bool = False,
                        max_tokens: Optional[int] = None,
                        model_params: Optional[dict] = None,
                        ) -> Union[ChatResponse, Iterable[StreamingChatChunk]]:
        pass

    @abstractmethod
    def enforced_function_call(self,
                               messages: Messages,
                               function: Function,
                               *,
                               max_tokens: Optional[int] = None,
                               model_params: Optional[dict] = None
                               ) -> dict:
        pass

    @abstractmethod
    async def achat_completion(self,
                               messages: Messages,
                               *,
                               stream: bool = False,
                               max_generated_tokens: Optional[int] = None,
                               model_params: Optional[dict] = None,
                               ) -> Union[ChatResponse,
                                          Iterable[StreamingChatChunk]]:
        pass

    @abstractmethod
    async def agenerate_queries(self,
                                messages: Messages,
                                *,
                                max_generated_tokens: Optional[int] = None,
                                model_params: Optional[dict] = None,
                                ) -> List[Query]:
        pass



import os
from abc import ABC, abstractmethod
from typing import Iterable, Union, Optional, cast

from canopy.chat_engine.models import HistoryPruningMethod
from canopy.chat_engine.prompt_builder import PromptBuilder
from canopy.chat_engine.query_generator import (QueryGenerator,
                                                FunctionCallingQueryGenerator, )
from canopy.context_engine import ContextEngine
from canopy.tokenizer import Tokenizer
from canopy.llm import BaseLLM, OpenAILLM
from canopy.models.api_models import (StreamingChatChunk, ChatResponse,
                                      StreamingChatResponse, )
from canopy.models.data_models import Context, Messages, SystemMessage
from canopy.utils.config import ConfigurableMixin

CE_DEBUG_INFO = os.getenv("CE_DEBUG_INFO", "FALSE").lower() == "true"


DEFAULT_SYSTEM_PROMPT = """Use the following pieces of context to answer the user question at the next messages. This context retrieved from a knowledge database and you should use only the facts from the context to answer. Always remember to include the source to the documents you used from their 'source' field in the format 'Source: $SOURCE_HERE'.
If you don't know the answer, just say that you don't know, don't try to make up an answer, use the context.
Don't address the context directly, but use it to answer the user question like it's your own knowledge.
"""  # noqa


class BaseChatEngine(ABC, ConfigurableMixin):
    @abstractmethod
    def chat(self,
             messages: Messages,
             *,
             stream: bool = False,
             model_params: Optional[dict] = None
             ) -> Union[ChatResponse, StreamingChatResponse]:
        pass

    @abstractmethod
    async def achat(self,
                    messages: Messages,
                    *,
                    stream: bool = False,
                    model_params: Optional[dict] = None
                    ) -> Union[ChatResponse, StreamingChatResponse]:
        pass

    @abstractmethod
    async def aget_context(self, messages: Messages) -> Context:
        pass


class ChatEngine(BaseChatEngine):

    """
    Chat engine is an object that implements end to end chat API with [RAG](https://www.pinecone.io/learn/retrieval-augmented-generation/).
    Given chat history, the chat engine orchestrates its underlying context engine and LLM to run the following steps:

    1. Generate search queries from the chat history
    2. Retrieve the most relevant context for each query using the context engine
    3. Prompt the LLM with the chat history and the retrieved context to generate the next response

    To use the chat engine, you need to provide it with a context engine, and optionally you can configure all its other components such as the LLM, the query generator, and the prompt builder and more.

    Example:

        >>> from canopy.chat_engine import ChatEngine
        >>> chat_engine = ChatEngine(context_engine)

    Where you can follow the instructions in the [context engine](../context_engine/context_engine) to create a context engine.
    Then you can use the chat engine to chat with a user:

        >>> from canopy.models.data_models import UserMessage
        >>> messages = [UserMessage(content="Hello! what is the capital of France?")]
        >>> response = chat_engine.chat(messages)
        >>> print(response.choices[0].message.content)
        Paris is the capital of France. Source: https://en.wikipedia.org/wiki/Paris
    """  # noqa: E501

    _DEFAULT_COMPONENTS = {
        'context_engine': ContextEngine,
        'llm': OpenAILLM,
        'query_builder': FunctionCallingQueryGenerator,
    }

    def __init__(self,
                 context_engine: ContextEngine,
                 *,
                 llm: Optional[BaseLLM] = None,
                 max_prompt_tokens: int = 4096,
                 max_generated_tokens: Optional[int] = None,
                 max_context_tokens: Optional[int] = None,
                 query_builder: Optional[QueryGenerator] = None,
                 system_prompt: Optional[str] = None,
                 history_pruning: str = "recent",
                 min_history_messages: int = 1
                 ):
        """
        Initialize a chat engine.

        Args:
            context_engine: An instance of a context engine to use for retrieving context to prompt the LLM along with the chat history.
            llm: An instance of a LLM to use for generating the next response. Defaults to OpenAILLM.
            max_prompt_tokens: The maximum number of tokens to use for the prompt to the LLM. Defaults to 4096.
            max_generated_tokens: The maximum number of tokens to generate from the LLM. Defaults to None, which means the LLM will use its default behavior.
            max_context_tokens: The maximum number of tokens to use for the context to prompt the LLM. Defaults to be 70% of the max_prompt_tokens.
            query_builder: An instance of a query generator to use for generating queries from the chat history. Defaults to FunctionCallingQueryGenerator.
            system_prompt: The system prompt to use for the LLM. Defaults to a generic prompt that is suitable for most use cases.
            history_pruning: The history pruning method to use for truncating the chat history to a prompt. Defaults to "recent", which means the chat history will be truncated to the most recent messages.
            min_history_messages: The minimum number of messages to keep in the chat history. Defaults to 1.
        """  # noqa: E501
        if not isinstance(context_engine, ContextEngine):
            raise TypeError(
                f"context_engine must be an instance of ContextEngine, "
                f"got {type(context_engine)}"
            )
        self.context_engine = context_engine

        if llm:
            if not isinstance(llm, BaseLLM):
                raise TypeError(
                    f"llm must be an instance of BaseLLM, got {type(llm)}"
                )
            self.llm = llm
        else:
            self.llm = self._DEFAULT_COMPONENTS['llm']()

        if query_builder:
            if not isinstance(query_builder, QueryGenerator):
                raise TypeError(
                    f"query_builder must be an instance of QueryGenerator, "
                    f"got {type(query_builder)}"
                )
            self._query_builder = query_builder
        else:
            self._query_builder = self._DEFAULT_COMPONENTS['query_builder']()

        self.max_prompt_tokens = max_prompt_tokens
        self.max_generated_tokens = max_generated_tokens
        self.system_prompt_template = system_prompt or DEFAULT_SYSTEM_PROMPT
        self._tokenizer = Tokenizer()
        self._prompt_builder = PromptBuilder(
            history_pruning=HistoryPruningMethod(history_pruning),
            min_history_messages=min_history_messages
        )

        # Set max budget for context tokens, default to 70% of max_prompt_tokens
        max_context_tokens = max_context_tokens or int(max_prompt_tokens * 0.7)
        system_prompt_tokens = self._tokenizer.messages_token_count(
            [SystemMessage(content=self.system_prompt_template)]
        )
        if max_context_tokens + system_prompt_tokens > max_prompt_tokens:
            raise ValueError(
                f"Not enough token budget for knowledge base context. The system prompt"
                f" is taking {system_prompt_tokens} tokens, and together with the "
                f"configured max context tokens {max_context_tokens} it exceeds "
                f"max_prompt_tokens of {self.max_prompt_tokens}"
            )
        self.max_context_tokens = max_context_tokens

    def chat(self,
             messages: Messages,
             *,
             stream: bool = False,
             model_params: Optional[dict] = None
             ) -> Union[ChatResponse, StreamingChatResponse]:
        """
        Chat completion with RAG. Given a list of messages (history), the chat engine will generate the next response, based on the relevant context retrieved from the knowledge base.

        While calling the chat method, behind the scenes the chat engine will do the following:
        1. Generate search queries from the chat history
        2. Retrieve the most relevant context for each query using the context engine
        3. Prompt the LLM with the chat history and the retrieved context to generate the next response
        4. Return the response

        Args:
            messages: A list of messages (history) to generate the next response from.
            stream: A boolean flag to indicate if the chat should be streamed or not. Defaults to False.
            model_params: A dictionary of model parameters to use for the LLM. Defaults to None, which means the LLM will use its default values.

        Returns:
            A ChatResponse object if stream is False, or a StreamingChatResponse object if stream is True.

        Examples:

            >>> from canopy.models.data_models import UserMessage
            >>> messages = [UserMessage(content="Hello! what is the capital of France?")]
            >>> response = chat_engine.chat(messages)
            >>> print(response.choices[0].message.content)

            Or you can stream the response:
            >>> response = chat_engine.chat(messages, stream=True)
            >>> for chunk in response.chunks:
            ...     print(chunk.json())
        """  # noqa: E501
        context = self._get_context(messages)
        system_prompt = self.system_prompt_template + f"\nContext: {context.to_text()}"
        llm_messages = self._prompt_builder.build(
            system_prompt,
            messages,
            max_tokens=self.max_prompt_tokens
        )
        llm_response = self.llm.chat_completion(llm_messages,
                                                max_tokens=self.max_generated_tokens,
                                                stream=stream,
                                                model_params=model_params)
        debug_info = {}
        if CE_DEBUG_INFO:
            debug_info['context'] = context.dict()
            debug_info['context'].update(context.debug_info)

        if stream:
            return StreamingChatResponse(
                chunks=cast(Iterable[StreamingChatChunk], llm_response),
                debug_info=debug_info
            )
        else:
            response = cast(ChatResponse, llm_response)
            response.debug_info = debug_info
            return response

    def _get_context(self,
                     messages: Messages,
                     ) -> Context:
        queries = self._query_builder.generate(messages, self.max_prompt_tokens)
        context = self.context_engine.query(queries, self.max_context_tokens)
        return context

    async def achat(self,
                    messages: Messages,
                    *,
                    stream: bool = False,
                    model_params: Optional[dict] = None
                    ) -> Union[ChatResponse, StreamingChatResponse]:
        raise NotImplementedError

    async def aget_context(self, messages: Messages) -> Context:
        raise NotImplementedError



import os
from abc import ABC, abstractmethod
from typing import List, Optional

from canopy.context_engine.context_builder import StuffingContextBuilder
from canopy.context_engine.context_builder.base import ContextBuilder
from canopy.knowledge_base import KnowledgeBase
from canopy.knowledge_base.base import BaseKnowledgeBase
from canopy.models.data_models import Context, Query
from canopy.utils.config import ConfigurableMixin

CE_DEBUG_INFO = os.getenv("CE_DEBUG_INFO", "FALSE").lower() == "true"


class BaseContextEngine(ABC, ConfigurableMixin):

    @abstractmethod
    def query(self, queries: List[Query], max_context_tokens: int, ) -> Context:
        pass

    @abstractmethod
    async def aquery(self, queries: List[Query], max_context_tokens: int, ) -> Context:
        pass


class ContextEngine(BaseContextEngine):
    """
    ContextEngine is responsible for providing context to the LLM, given a set of search queries.

    Once called with a set of queries, the ContextEngine will go through the following steps:
    1. Query the knowledge base for relevant documents
    2. Build a context from the documents retrieved that can be injected into the LLM prompt

    The context engine considers token budgeting when building the context, and tries to maximize the amount of relevant information that can be provided to the LLM within the token budget.

    To create a context engine, you must provide a knowledge base and optionally a context builder.

    Example:
    >>> from canopy.context_engine import ContextEngine
    >>> from canopy.models.data_models import Query
    >>> context_engine = ContextEngine(knowledge_base=knowledge_base)
    >>> context_engine.query(Query(text="What is the capital of France?"), max_context_tokens=1000)

    To create a knowledge base, see the documentation for the knowledge base module (canopy.knowledge_base.knowledge_base).
    """  # noqa: E501

    _DEFAULT_COMPONENTS = {
        'knowledge_base': KnowledgeBase,
        'context_builder': StuffingContextBuilder,
    }

    def __init__(self,
                 knowledge_base: BaseKnowledgeBase,
                 *,
                 context_builder: Optional[ContextBuilder] = None,
                 global_metadata_filter: Optional[dict] = None
                 ):
        """
        Initialize a new ContextEngine.

        Args:
            knowledge_base: The knowledge base to query for retrieving documents
            context_builder: The context builder to use for building the context from the retrieved documents. Defaults to `StuffingContextBuilder`
            global_metadata_filter: A metadata filter to apply to all queries. See: https://docs.pinecone.io/docs/metadata-filtering
        """  # noqa: E501

        if not isinstance(knowledge_base, BaseKnowledgeBase):
            raise TypeError("knowledge_base must be an instance of BaseKnowledgeBase, "
                            f"not {type(self.knowledge_base)}")
        self.knowledge_base = knowledge_base

        if context_builder:
            if not isinstance(context_builder, ContextBuilder):
                raise TypeError(
                    "context_builder must be an instance of ContextBuilder, "
                    f"not {type(context_builder)}"
                )
            self.context_builder = context_builder
        else:
            self.context_builder = self._DEFAULT_COMPONENTS['context_builder']()

        self.global_metadata_filter = global_metadata_filter

    def query(self, queries: List[Query], max_context_tokens: int, ) -> Context:
        """
        Query the knowledge base for relevant documents and build a context from the retrieved documents that can be injected into the LLM prompt.

        Args:
            queries: A list of queries to use for retrieving documents from the knowledge base
            max_context_tokens: The maximum number of tokens to use for the context

        Returns:
            A Context object containing the retrieved documents and metadata

        Example:
        >>> from canopy.context_engine import ContextEngine
        >>> from canopy.models.data_models import Query
        >>> context_engine = ContextEngine(knowledge_base=knowledge_base)
        >>> context_engine.query(Query(text="What is the capital of France?"), max_context_tokens=1000)
        """  # noqa: E501
        query_results = self.knowledge_base.query(
            queries,
            global_metadata_filter=self.global_metadata_filter)
        context = self.context_builder.build(query_results, max_context_tokens)

        if CE_DEBUG_INFO:
            context.debug_info["query_results"] = [qr.dict() for qr in query_results]
        return context

    async def aquery(self, queries: List[Query], max_context_tokens: int, ) -> Context:
        raise NotImplementedError()



from abc import ABC, abstractmethod
from typing import List, Optional

from canopy.knowledge_base.models import KBEncodedDocChunk, KBQuery, KBDocChunk
from canopy.models.data_models import Query
from canopy.utils.config import ConfigurableMixin


class RecordEncoder(ABC, ConfigurableMixin):
    """
    Base class for RecordEncoders. Encodes document chunks and queries to vector representations.
    The vector representation may include both dense and sparse values.
    Dense values are usually generated by an embedding model, and sparse values usually represent weighted keyword counts.

    The RecordEncoder implements separate functions for the encoding of documents and queries.
    Some implementations of both sparse and dense encoding are not symmetrical. For example, BM25 sparse
    encoders and instruction dense encoders.

    Any class that extends RecordEncoder must implement the methods for encoding documents and queries:
    - _encode_documents_batch
    - _encode_queries_batch

    Async encoders are still not supported, but will be added in the future.
    """   # noqa: E501

    def __init__(self, batch_size: int = 1):
        """
        Initialize the encoder.

        Args:
            batch_size: The number of documents or queries to encode at once.
                        Defaults to 1.
        """   # noqa: E501
        self.batch_size = batch_size

    # TODO: rename documents to doc_chunks or chunks
    @abstractmethod
    def _encode_documents_batch(self,
                                documents: List[KBDocChunk]
                                ) -> List[KBEncodedDocChunk]:
        """
        Abstract method for encoding a batch of documents, takes a list of KBDocChunk and returns a list of KBEncodedDocChunk.
        For maximal performance, and derived class should try to operate on the entire documents batch in a single operation.

        Args:
            documents: A list of KBDocChunk to encode.

        Returns:
            encoded chunks: A list of KBEncodedDocChunk.
        """   # noqa: E501
        pass

    @abstractmethod
    def _encode_queries_batch(self, queries: List[Query]) -> List[KBQuery]:
        """
        Abstract method for encoding a batch of queries, takes a list of Query and returns a list of KBQuery.
        For maximal performance, and derived class should try to operate on the entire batch in a single operation.

        Args:
            queries: A list of `Query` objects to encode.

        Returns:
            encoded queries: A list of KBQuery.
        """   # noqa: E501
        pass

    @abstractmethod
    async def _aencode_documents_batch(self,
                                       documents: List[KBDocChunk]
                                       ) -> List[KBEncodedDocChunk]:
        raise NotImplementedError

    @abstractmethod
    async def _aencode_queries_batch(self, queries: List[Query]) -> List[KBQuery]:
        raise NotImplementedError

    @staticmethod
    def _batch_iterator(data: list, batch_size):
        return (data[pos:pos + batch_size] for pos in range(0, len(data), batch_size))

    @property
    def dimension(self) -> Optional[int]:
        """
        Returns:
            The dimension of the dense vectors produced by the encoder, if applicable.
        """   # noqa: E501
        return None

    def encode_documents(self, documents: List[KBDocChunk]) -> List[KBEncodedDocChunk]:
        """

        Encode documents in batches. Will iterate over batch of documents and encode them using the _encode_documents_batch method.

        Args:
            documents: A list of KBDocChunk to encode.

        Returns:
            encoded chunks: A list of KBEncodedDocChunk.

        """   # noqa: E501
        encoded_docs = []
        for batch in self._batch_iterator(documents, self.batch_size):
            try:
                encoded_docs.extend(self._encode_documents_batch(batch))
            except Exception as e:
                raise RuntimeError(
                    f"Failed to enconde documents using {self.__class__.__name__}. "
                    f"Error: {self._format_error(e)}"
                ) from e

        return encoded_docs  # TODO: consider yielding a generator

    def encode_queries(self, queries: List[Query]) -> List[KBQuery]:
        """

        Encode queries in batches. Will iterate over batch of queries and encode them using the _encode_queries_batch method.

        Args:
            queries: A list of Query to encode.

        Returns:
            encoded queries: A list of KBQuery.
        """   # noqa: E501

        kb_queries = []
        for batch in self._batch_iterator(queries, self.batch_size):
            try:
                kb_queries.extend(self._encode_queries_batch(batch))
            except Exception as e:
                raise RuntimeError(
                    f"Failed to enconde queries using {self.__class__.__name__}. "
                    f"Error: {self._format_error(e)}"
                ) from e

        return kb_queries

    async def aencode_documents(self,
                                documents: List[KBDocChunk]
                                ) -> List[KBEncodedDocChunk]:
        encoded_docs = []
        for batch in self._batch_iterator(documents, self.batch_size):
            encoded_docs.extend(await self._aencode_documents_batch(batch))

        return encoded_docs

    async def aencode_queries(self, queries: List[Query]) -> List[KBQuery]:
        kb_queries = []
        for batch in self._batch_iterator(queries, self.batch_size):
            kb_queries.extend(await self._aencode_queries_batch(batch))

        return kb_queries

    def _format_error(self, err):
        return f"{err}"



from copy import deepcopy
from typing import List, Optional

from .langchain_text_splitter import RecursiveCharacterTextSplitter

from canopy.knowledge_base.chunker.base import Chunker
from canopy.knowledge_base.models import KBDocChunk
from canopy.tokenizer import Tokenizer
from canopy.models.data_models import Document


class RecursiveCharacterChunker(Chunker):
    """
    A chunker that splits a document into chunks of a given size, using a recursive character splitter.
    A RecursiveCharacterChunker is a derived class of Chunker, which means that it can be referenced by a name
    and configured in a config file.
    """  # noqa: E501

    def __init__(self,
                 chunk_size: int = 256,
                 chunk_overlap: int = 0,
                 separators: Optional[List[str]] = None,
                 keep_separator: bool = True,
                 ):
        """
        RecursiveCharacterTextSplitter is a text splitter from the langchain library.
        It splits a text into chunks of a given size, using a recursive character splitter.

        Args:
            chunk_size: size of the chunks, in tokens
            chunk_overlap: overlap between chunks
            separators: list of separators to use for splitting the text
            keep_separator: whether to keep the separator in the chunk or not
        """  # noqa: E501
        self._tokenizer = Tokenizer()
        self._chunker = RecursiveCharacterTextSplitter(
            chunk_size=chunk_size,
            chunk_overlap=chunk_overlap,
            length_function=self._tokenizer.token_count,
            separators=separators,
            keep_separator=keep_separator)

    def chunk_single_document(self, document: Document) -> List[KBDocChunk]:
        """
        using the RecursiveCharacterTextSplitter, this method takes a document and returns a list of KBDocChunks
        Args:
            document: document to be chunked

        Returns:
            chunks: list of chunks KBDocChunks from the document, where text is splitted
                              evenly using the RecursiveCharacterTextSplitter
        """  # noqa: E501
        # TODO: check overlap not bigger than max_chunk_size
        text_chunks = self._chunker.split_text(document.text)
        return [KBDocChunk(id=f"{document.id}_{i}",
                           document_id=document.id,
                           text=text_chunk,
                           source=document.source,
                           metadata=deepcopy(document.metadata))
                for i, text_chunk in enumerate(text_chunks)]

    async def achunk_single_document(self, document: Document) -> List[KBDocChunk]:
        raise NotImplementedError()



from typing import List
from pinecone_text.dense.cohere_encoder import CohereEncoder
from canopy.knowledge_base.models import KBDocChunk, KBEncodedDocChunk, KBQuery
from canopy.knowledge_base.record_encoder.dense import DenseRecordEncoder
from canopy.models.data_models import Query


class CohereRecordEncoder(DenseRecordEncoder):
    """
    CohereRecordEncoder is a type of DenseRecordEncoder that uses the Cohere `embed` API.
    The implementation uses the `CohereEncoder` class from the `pinecone-text` library.
    For more information about see: https://github.com/pinecone-io/pinecone-text

    """  # noqa: E501

    def __init__(
        self,
        *,
        model_name: str = "embed-english-v3.0",
        batch_size: int = 100,
        **kwargs,
    ):
        """
        Initialize the CohereRecordEncoder

        Args:
            model_name: The name of the Cohere embeddings model to use for encoding. See https://docs.cohere.com/reference/embed
            batch_size: The number of documents or queries to encode at once.
                        Defaults to 400.
            **kwargs: Additional arguments to pass to the underlying `pinecone-text. CohereEncoder`.
        """  # noqa: E501
        encoder = CohereEncoder(model_name, **kwargs)
        super().__init__(dense_encoder=encoder, batch_size=batch_size)

    def encode_documents(self, documents: List[KBDocChunk]) -> List[KBEncodedDocChunk]:
        """
        Encode a list of documents, takes a list of KBDocChunk and returns a list of KBEncodedDocChunk.

        Args:
            documents: A list of KBDocChunk to encode.

        Returns:
            encoded chunks: A list of KBEncodedDocChunk, with the `values` field populated by the generated embeddings vector.
        """  # noqa: E501
        return super().encode_documents(documents)

    async def _aencode_documents_batch(
        self, documents: List[KBDocChunk]
    ) -> List[KBEncodedDocChunk]:
        raise NotImplementedError

    async def _aencode_queries_batch(self, queries: List[Query]) -> List[KBQuery]:
        raise NotImplementedError



from abc import ABC, abstractmethod
from typing import List, Optional

from canopy.knowledge_base.models import QueryResult
from canopy.models.data_models import Query, Document
from canopy.utils.config import ConfigurableMixin


class BaseKnowledgeBase(ABC, ConfigurableMixin):
    """
    KnowledgeBase is an abstract class that defines the interface for a knowledge base.
    """

    @abstractmethod
    def query(self,
              queries: List[Query],
              global_metadata_filter: Optional[dict] = None
              ) -> List[QueryResult]:
        pass

    @abstractmethod
    def upsert(self,
               documents: List[Document],
               namespace: str = "", ) -> None:
        pass

    # TODO: Do we want delete by metadata?
    @abstractmethod
    def delete(self,
               document_ids: List[str],
               namespace: str = "") -> None:
        pass

    @abstractmethod
    def verify_index_connection(self) -> None:
        pass

    @abstractmethod
    async def aquery(self,
                     queries: List[Query],
                     global_metadata_filter: Optional[dict] = None
                     ) -> List[QueryResult]:
        pass

    @abstractmethod
    async def aupsert(self,
                      documents: List[Document],
                      namespace: str = "",
                      ) -> None:
        pass

    @abstractmethod
    async def adelete(self,
                      document_ids: List[str],
                      namespace: str = "") -> None:
        pass



from abc import ABC, abstractmethod
from typing import List

from canopy.models.data_models import Messages, Query
from canopy.utils.config import ConfigurableMixin


class QueryGenerator(ABC, ConfigurableMixin):
    @abstractmethod
    def generate(self,
                 messages: Messages,
                 max_prompt_tokens: int,
                 ) -> List[Query]:
        pass

    @abstractmethod
    async def agenerate(self,
                        messages: Messages,
                        max_prompt_tokens: int,
                        ) -> List[Query]:
        raise NotImplementedError



import os
from copy import deepcopy
from datetime import datetime
import time
from typing import List, Optional, Dict, Any
import pandas as pd
from pinecone import list_indexes, delete_index, create_index, init \
    as pinecone_init, whoami as pinecone_whoami
from pinecone import ApiException as PineconeApiException

try:
    from pinecone import GRPCIndex as Index
except ImportError:
    from pinecone import Index

from pinecone_datasets import Dataset
from pinecone_datasets import DenseModelMetadata, DatasetMetadata

from canopy.knowledge_base.base import BaseKnowledgeBase
from canopy.knowledge_base.chunker import Chunker, MarkdownChunker
from canopy.knowledge_base.record_encoder import (RecordEncoder,
                                                  OpenAIRecordEncoder)
from canopy.knowledge_base.models import (KBQueryResult, KBQuery, QueryResult,
                                          KBDocChunkWithScore, DocumentWithScore)
from canopy.knowledge_base.reranker import Reranker, TransparentReranker
from canopy.models.data_models import Query, Document


INDEX_NAME_PREFIX = "canopy--"
TIMEOUT_INDEX_CREATE = 300
TIMEOUT_INDEX_PROVISION = 30
INDEX_PROVISION_TIME_INTERVAL = 3
RESERVED_METADATA_KEYS = {"document_id", "text", "source"}

DELETE_STARTER_BATCH_SIZE = 30

DELETE_STARTER_CHUNKS_PER_DOC = 32


def connect_to_pinecone():
    """
    Connect to Pinecone.
    This method is called automatically when creating a new KnowledgeBase object.
    Or when calling `list_canopy_indexes()`.
    """
    try:
        pinecone_init()
        pinecone_whoami()
    except Exception as e:
        raise RuntimeError("Failed to connect to Pinecone. "
                           "Please check your credentials and try again") from e


def list_canopy_indexes() -> List[str]:
    """
    List all Canopy indexes in the current Pinecone account.

    Example:
        >>> from canopy.knowledge_base import list_canopy_indexes
        >>> list_canopy_indexes()
            ['canopy--my_index', 'canopy--my_index2']

    Returns:
        A list of Canopy index names.
    """

    connect_to_pinecone()
    return [index for index in list_indexes() if index.startswith(INDEX_NAME_PREFIX)]


class KnowledgeBase(BaseKnowledgeBase):

    """
    The `KnowledgeBase` is used to store and retrieve text documents, using an underlying Pinecone index.
    Every document is chunked into multiple text snippets based on the text structure (e.g. Markdown or HTML formatting)
    Then, each chunk is encoded into a vector using an embedding model, and the resulting vectors are inserted to the Pinecone index.
    After documents were inserted, the KnowledgeBase can be queried by sending a textual query, which will first encoded to a vector
    and then used to retrieve the closest top-k document chunks.

    Note: Since Canopy defines its own data format, you can not use a pre-existing Pinecone index with Canopy's KnowledgeBase.
          The index must be created by using `knowledge_base.create_canopy_index()` or the CLI command `canopy new`.

    When creating a new Canopy service, the user must first create the underlying Pinecone index.
    This is a one-time setup process - the index will exist on Pinecone's managed service until it is deleted.

    Example:
        >>> from canopy.knowledge_base.knowledge_base import KnowledgeBase
        >>> from tokenizer import Tokenizer
        >>> Tokenizer.initialize()
        >>> kb = KnowledgeBase(index_name="my_index")
        >>> kb.create_canopy_index()

    In any future interactions, the user simply needs to connect to the existing index:

        >>> kb = KnowledgeBase(index_name="my_index")
        >>> kb.connect()
    """  # noqa: E501

    _DEFAULT_COMPONENTS = {
        'record_encoder': OpenAIRecordEncoder,
        'chunker': MarkdownChunker,
        'reranker': TransparentReranker
    }

    def __init__(self,
                 index_name: str,
                 *,
                 record_encoder: Optional[RecordEncoder] = None,
                 chunker: Optional[Chunker] = None,
                 reranker: Optional[Reranker] = None,
                 default_top_k: int = 5,
                 ):
        """
        Initilize the knowledge base object.

        If the index does not exist, the user must first create it by calling `create_canopy_index()` or the CLI command `canopy new`.

        Note: Canopy will add the prefix --canopy to your selected index name.
             You can retrieve the full index name knowledge_base.index_name at any time, or find it in the Pinecone console at https://app.pinecone.io/

        Example:

            create a new index:
            >>> from canopy.knowledge_base.knowledge_base import KnowledgeBase
            >>> from tokenizer import Tokenizer
            >>> Tokenizer.initialize()
            >>> kb = KnowledgeBase(index_name="my_index")
            >>> kb.create_canopy_index()

        In any future interactions,
        the user simply needs to connect to the existing index:

            >>> kb = KnowledgeBase(index_name="my_index")
            >>> kb.connect()

        Args:
            index_name: The name of the underlying Pinecone index.
            record_encoder: An instance of RecordEncoder to use for encoding documents and queries.
                                                      Defaults to OpenAIRecordEncoder.
            chunker: An instance of Chunker to use for chunking documents. Defaults to MarkdownChunker.
            reranker: An instance of Reranker to use for reranking query results. Defaults to TransparentReranker.
            default_top_k: The default number of document chunks to return per query. Defaults to 5.

        Raises:
            ValueError: If default_top_k is not a positive integer.
            TypeError: If record_encoder is not an instance of RecordEncoder.
            TypeError: If chunker is not an instance of Chunker.
            TypeError: If reranker is not an instance of Reranker.

        """  # noqa: E501
        if default_top_k < 1:
            raise ValueError("default_top_k must be greater than 0")

        self._index_name = self._get_full_index_name(index_name)
        self._default_top_k = default_top_k

        if record_encoder:
            if not isinstance(record_encoder, RecordEncoder):
                raise TypeError(
                    f"record_encoder must be an instance of RecordEncoder, "
                    f"not {type(record_encoder)}"
                )
            self._encoder = record_encoder
        else:
            self._encoder = self._DEFAULT_COMPONENTS['record_encoder']()

        if chunker:
            if not isinstance(chunker, Chunker):
                raise TypeError(
                    f"chunker must be an instance of Chunker, not {type(chunker)}"
                )
            self._chunker = chunker
        else:
            self._chunker = self._DEFAULT_COMPONENTS['chunker']()

        if reranker:
            if not isinstance(reranker, Reranker):
                raise TypeError(
                    f"reranker must be an instance of Reranker, not {type(reranker)}"
                )
            self._reranker = reranker
        else:
            self._reranker = self._DEFAULT_COMPONENTS['reranker']()

        # Normally, index creation params are passed directly to the `.create_canopy_index()` method.  # noqa: E501
        # However, when KnowledgeBase is initialized from a config file, these params
        # would be set by the `KnowledgeBase.from_config()` constructor.
        self._index_params: Dict[str, Any] = {}

        # The index object is initialized lazily, when the user calls `connect()` or
        # `create_canopy_index()`
        self._index: Optional[Index] = None

    def _connect_index(self,
                       connect_pinecone: bool = True
                       ) -> None:
        if connect_pinecone:
            connect_to_pinecone()

        if self.index_name not in list_indexes():
            raise RuntimeError(
                f"The index {self.index_name} does not exist or was deleted. "
                "Please create it by calling knowledge_base.create_canopy_index() or "
                "running the `canopy new` command"
            )

        try:
            self._index = Index(index_name=self.index_name)
            self.verify_index_connection()
        except Exception as e:
            self._index = None
            raise RuntimeError(
                f"Unexpected error while connecting to index {self.index_name}. "
                f"Please check your credentials and try again."
            ) from e

    @property
    def _connection_error_msg(self) -> str:
        return (
            f"KnowledgeBase is not connected to index {self.index_name}, "
            f"Please call knowledge_base.connect(). "
        )

    def connect(self) -> None:
        """
        Connect to the underlying Pinecone index.
        This method must be called before making any other calls to the knowledge base.

        Note: If underlying index is not provisioned yet, an exception will be raised.
              To provision the index, use `create_canopy_index()` or the CLI command `canopy new`.

        Returns:
            None

        Raises:
            RuntimeError: If the knowledge base failed to connect to the underlying Pinecone index.
        """  # noqa: E501
        if self._index is None:
            self._connect_index()

    def verify_index_connection(self) -> None:
        """
        Verify that the knowledge base is connected to the underlying Pinecone index.

        Returns:
            None

        Raises:
            RuntimeError: If the knowledge base is not connected properly to the underlying Pinecone index.
        """  # noqa: E501
        if self._index is None:
            raise RuntimeError(self._connection_error_msg)

        try:
            self._index.describe_index_stats()
        except Exception as e:
            raise RuntimeError(
                "The index did not respond. Please check your credentials and try again"
            ) from e

    def create_canopy_index(self,
                            indexed_fields: Optional[List[str]] = None,
                            dimension: Optional[int] = None,
                            index_params: Optional[dict] = None
                            ):
        """
        Creates the underlying Pinecone index that will be used by the KnowledgeBase.
        This is a one time set-up operation that only needs to be done once for every new Canopy service.
        After the index was created, it will persist in Pinecone until explicitly deleted.

        Since Canopy defines its own data format, namely a few dedicated metadata fields,
        you can not use a pre-existing Pinecone index with Canopy's KnowledgeBase.
        The index must be created by using `knowledge_base.create_canopy_index()` or the CLI command `canopy new`.

        Note: This operation may take a few minutes to complete.
              Once created, you can see the index in the Pinecone console

        Note: Canopy will add the prefix --canopy to your selected index name.
             You can retrieve the full index name knowledge_base.index_name at any time.
             Or find it in the Pinecone console at https://app.pinecone.io/

        Args:
           indexed_fields: A list of metadata fields that would be indexed,
                           allowing them to be later used for metadata filtering.
                           All other metadata fields are stored but not indexed.
                           See: https://docs.pinecone.io/docs/manage-indexes#selective-metadata-indexing.
                           Canopy always indexes a built-in `document_id` field, which is added to every vector.
                           By default - all other metadata fields are **not** indexed, unless explicitly defined in this list.
           dimension: The dimension of the vectors to index.
                       If `dimension` isn't explicitly provided,
                       Canopy would try to infer the embedding's dimension based on the configured `Encoder`
           index_params: A dictionary of parameters to pass to the index creation API.
                         For example, you can set the index's number of replicas by passing {"replicas": 2}.
                         see https://docs.pinecone.io/docs/python-client#create_index

        """  # noqa: E501
        # validate inputs
        if indexed_fields is None:
            indexed_fields = ['document_id']
        elif "document_id" not in indexed_fields:
            indexed_fields.append('document_id')

        if 'text' in indexed_fields:
            raise ValueError("The 'text' field cannot be used for metadata filtering. "
                             "Please remove it from indexed_fields")

        if dimension is None:
            try:
                encoder_dimension = self._encoder.dimension
                if encoder_dimension is None:
                    raise RuntimeError(
                        f"The selected encoder {self._encoder.__class__.__name__} does "
                        f"not support inferring the vectors' dimensionality."
                    )
                dimension = encoder_dimension
            except Exception as e:
                raise RuntimeError(
                    f"Canopy has failed to infer vectors' dimensionality using the "
                    f"selected encoder: {self._encoder.__class__.__name__}. You can "
                    f"provide the dimension manually, try using a different encoder, or"
                    f" fix the underlying error:\n{e}"
                ) from e

        # connect to pinecone and create index
        connect_to_pinecone()

        if self.index_name in list_indexes():
            raise RuntimeError(
                f"Index {self.index_name} already exists. To connect to an "
                f"existing index, use `knowledge_base.connect()`. "
                "If you wish to delete it call `knowledge_base.delete_index()`. "
            )

        # create index
        index_params = index_params or self._index_params
        try:
            create_index(name=self.index_name,
                         dimension=dimension,
                         metadata_config={
                             'indexed': indexed_fields
                         },
                         timeout=TIMEOUT_INDEX_CREATE,
                         **index_params)
        except (Exception, PineconeApiException) as e:
            raise RuntimeError(
                f"Failed to create index {self.index_name} due to error: "
                f"{e.body if isinstance(e, PineconeApiException) else e}"
            ) from e

        # wait for index to be provisioned
        self._wait_for_index_provision()

    def _wait_for_index_provision(self):
        start_time = time.time()
        while True:
            try:
                self._connect_index(connect_pinecone=False)
                break
            except RuntimeError:
                pass

            time_passed = time.time() - start_time
            if time_passed > TIMEOUT_INDEX_PROVISION:
                raise RuntimeError(
                    f"Index {self.index_name} failed to provision "
                    f"for {time_passed} seconds."
                    f"Please try creating KnowledgeBase again in a few minutes."
                )
            time.sleep(INDEX_PROVISION_TIME_INTERVAL)

    @staticmethod
    def _get_full_index_name(index_name: str) -> str:
        if index_name.startswith(INDEX_NAME_PREFIX):
            return index_name
        else:
            return INDEX_NAME_PREFIX + index_name

    @property
    def index_name(self) -> str:
        """
        The name of the index the knowledge base is connected to.
        """
        return self._index_name

    def delete_index(self):
        """
        Delete the underlying Pinecone index.

        **Note: THIS OPERATION IS NOT REVERSIBLE!!**
        Once deleted, the index together with any stored documents cannot be restored!

        After deletion - the `KnowledgeBase` would not be connected to a Pinecone index anymore,
                         so you will not be able to insert documents or query.
                         If you'd wish to re-create an index with the same name, simply call `knowledge_base.create_canopy_index()`
                         Or use the CLI command `canopy new`.
        """  # noqa: E501
        if self._index is None:
            raise RuntimeError(self._connection_error_msg)
        delete_index(self._index_name)
        self._index = None

    def query(self,
              queries: List[Query],
              global_metadata_filter: Optional[dict] = None
              ) -> List[QueryResult]:
        """
        Query the knowledge base to retrieve document chunks.

        This operation includes several steps:
        1. Encode the queries to vectors using the underlying encoder.
        2. Query the underlying Pinecone index to retrieve the top-k chunks for each query.
        3. Rerank the results using the underlying reranker.
        4. Return the results for each query as a list of QueryResult objects.

        Args:
            queries: A list of queries to run against the knowledge base.
            global_metadata_filter: A metadata filter to apply to all queries, in addition to any query-specific filters.
                                    For example, the filter {"website": "wiki"} will only return documents with the metadata {"website": "wiki"} (in case provided in upsert)
                                    see https://docs.pinecone.io/docs/metadata-filtering
        Returns:
            A list of QueryResult objects.

        Examples:
            >>> from canopy.knowledge_base.knowledge_base import KnowledgeBase
            >>> from tokenizer import Tokenizer
            >>> Tokenizer.initialize()
            >>> kb = KnowledgeBase(index_name="my_index")
            >>> kb.connect()
            >>> queries = [Query(text="How to make a cake"),
                           Query(text="How to make a pizza",
                                top_k=10,
                                metadata_filter={"website": "wiki"})]
            >>> results = kb.query(queries)
        """  # noqa: E501
        if self._index is None:
            raise RuntimeError(self._connection_error_msg)

        queries = self._encoder.encode_queries(queries)
        results = [self._query_index(q, global_metadata_filter) for q in queries]
        results = self._reranker.rerank(results)

        return [
            QueryResult(
                query=r.query,
                documents=[
                    DocumentWithScore(
                        **d.dict(exclude={
                            'values', 'sparse_values', 'document_id'
                        })
                    )
                    for d in r.documents
                ]
            ) for r in results
        ]

    def _query_index(self,
                     query: KBQuery,
                     global_metadata_filter: Optional[dict]) -> KBQueryResult:
        if self._index is None:
            raise RuntimeError(self._connection_error_msg)

        metadata_filter = deepcopy(query.metadata_filter)
        if global_metadata_filter is not None:
            if metadata_filter is None:
                metadata_filter = {}
            metadata_filter.update(global_metadata_filter)
        top_k = query.top_k if query.top_k else self._default_top_k

        query_params = deepcopy(query.query_params)
        _check_return_type = query.query_params.pop('_check_return_type', False)
        result = self._index.query(vector=query.values,
                                   sparse_vector=query.sparse_values,
                                   top_k=top_k,
                                   namespace=query.namespace,
                                   filter=metadata_filter,
                                   include_metadata=True,
                                   _check_return_type=_check_return_type,
                                   **query_params)
        documents: List[KBDocChunkWithScore] = []
        for match in result['matches']:
            metadata = match['metadata']
            text = metadata.pop('text')
            document_id = metadata.pop('document_id')
            documents.append(
                KBDocChunkWithScore(id=match['id'],
                                    text=text,
                                    document_id=document_id,
                                    score=match['score'],
                                    source=metadata.pop('source', ''),
                                    metadata=metadata)
            )
        return KBQueryResult(query=query.text, documents=documents)

    def upsert(self,
               documents: List[Document],
               namespace: str = "",
               batch_size: int = 200,
               show_progress_bar: bool = False):
        """
        Upsert documents into the knowledge base.
        Upsert operation stands for "update or insert".
        It means that if a document with the same id already exists in the index, it will be updated with the new document.
        Otherwise, a new document will be inserted.

        This operation includes several steps:
        1. Split the documents into smaller chunks.
        2. Encode the chunks to vectors.
        3. Delete any existing chunks belonging to the same documents.
        4. Upsert the chunks to the index.

        Args:
            documents: A list of documents to upsert.
            namespace: The namespace in the underlying index to upsert documents into.
            batch_size: Refers only to the actual upsert operation to the underlying index.
                        The number of chunks (multiple piecies of text per document) to upsert in each batch.
                        Defaults to 100.
            show_progress_bar: Whether to show a progress bar while upserting the documents.


        Example:
            >>> from canopy.knowledge_base.knowledge_base import KnowledgeBase
            >>> from tokenizer import Tokenizer
            >>> Tokenizer.initialize()
            >>> kb = KnowledgeBase(index_name="my_index")
            >>> kb.connect()
            >>> documents = [Document(id="doc1",
                                        text="This is a document",
                                        source="my_source",
                                        metadata={"website": "wiki"}),
                            Document(id="doc2",
                                     text="This is another document",
                                     source="my_source",
                                     metadata={"website": "wiki"})]
            >>> kb.upsert(documents)
        """  # noqa: E501
        if self._index is None:
            raise RuntimeError(self._connection_error_msg)

        for doc in documents:
            metadata_keys = set(doc.metadata.keys())
            forbidden_keys = metadata_keys.intersection(RESERVED_METADATA_KEYS)
            if forbidden_keys:
                raise ValueError(
                    f"Document with id {doc.id} contains reserved metadata keys: "
                    f"{forbidden_keys}. Please remove them and try again."
                )

        chunks = self._chunker.chunk_documents(documents)
        encoded_chunks = self._encoder.encode_documents(chunks)

        encoder_name = self._encoder.__class__.__name__

        dataset_metadata = DatasetMetadata(name=self._index_name,
                                           created_at=str(datetime.now()),
                                           documents=len(chunks),
                                           dense_model=DenseModelMetadata(
                                               name=encoder_name,
                                               dimension=self._encoder.dimension),
                                           queries=0)

        dataset = Dataset.from_pandas(
            pd.DataFrame.from_records([c.to_db_record() for c in encoded_chunks]),
            metadata=dataset_metadata
        )

        # The upsert operation may update documents which may already exist
        # int the index, as many individual chunks.
        # As the process of chunking might have changed
        # the number of chunks per document,
        # we need to delete all existing chunks
        # belonging to the same documents before upserting the new ones.
        # we currently don't delete documents before upsert in starter env
        if not self._is_starter_env():
            self.delete(document_ids=[doc.id for doc in documents],
                        namespace=namespace)

        # Upsert to Pinecone index
        dataset.to_pinecone_index(self._index_name,
                                  namespace=namespace,
                                  should_create_index=False,
                                  batch_size=batch_size,
                                  show_progress=show_progress_bar)

    def delete(self,
               document_ids: List[str],
               namespace: str = "") -> None:
        """
        Delete documents from the underlying Pinecone index.
        Since each document is chunked into multiple chunks, this operation will delete all chunks belonging to the given document ids.
        This operation not raise an exception if the document does not exist.

        Args:
            document_ids: A list of document ids to delete from the index.
            namespace: The namespace in the underlying index to delete documents from.

        Returns:
            None

        Example:
            >>> from canopy.knowledge_base.knowledge_base import KnowledgeBase
            >>> from tokenizer import Tokenizer
            >>> Tokenizer.initialize()
            >>> kb = KnowledgeBase(index_name="my_index")
            >>> kb.connect()
            >>> kb.delete(document_ids=["doc1", "doc2"])
        """  # noqa: E501
        if self._index is None:
            raise RuntimeError(self._connection_error_msg)

        # Currently starter env does not support delete by metadata filter
        # So temporarily we delete the first DELETE_STARTER_CHUNKS_PER_DOC chunks
        if self._is_starter_env():
            for i in range(0, len(document_ids), DELETE_STARTER_BATCH_SIZE):
                doc_ids_chunk = document_ids[i:i + DELETE_STARTER_BATCH_SIZE]
                chunked_ids = [f"{doc_id}_{i}"
                               for doc_id in doc_ids_chunk
                               for i in range(DELETE_STARTER_CHUNKS_PER_DOC)]
                try:
                    self._index.delete(ids=chunked_ids,
                                       namespace=namespace)
                except Exception as e:
                    raise RuntimeError(
                        f"Failed to delete document ids: {document_ids[i:]}"
                        f"Please try again."
                    ) from e
        else:
            self._index.delete(
                filter={"document_id": {"$in": document_ids}},
                namespace=namespace
            )

    @classmethod
    def from_config(cls,
                    config: Dict[str, Any],
                    index_name: Optional[str] = None) -> 'KnowledgeBase':
        """
        Create a KnowledgeBase object from a configuration dictionary.

        Args:
            config: A dictionary containing the configuration for the knowledge base.
            index_name: The name of the index to connect to (optional).
                        If not provided, the index name will be read from the environment variable INDEX_NAME.

        Returns:
            A KnowledgeBase object.
        """  # noqa: E501
        index_name = index_name or os.getenv("INDEX_NAME")
        if index_name is None:
            raise ValueError(
                "index_name must be provided. Either pass it explicitly or set the "
                "INDEX_NAME environment variable"
            )
        config = deepcopy(config)
        config['params'] = config.get('params', {})

        # Check if the config includes an 'index_name', which is not the same as the
        # index_name passed as argument \ environment variable.
        if config['params'].get('index_name', index_name) != index_name:
            raise ValueError(
                f"index_name in config ({config['params']['index_name']}), while "
                f"INDEX_NAME environment variable is {index_name}. "
                f"Please make sure they are the same or remove the 'index_name' key "
                f"from the config."
            )
        config['params']['index_name'] = index_name

        # If the config includes an 'index_params' key, they need to be saved until
        # the index is created, and then passed to the index creation method.
        index_params = config['params'].pop('index_params', {})
        kb = cls._from_config(config)
        kb._index_params = index_params
        return kb

    @staticmethod
    def _is_starter_env():
        starter_env_suffixes = ("starter", "stage-gcp-0")
        return os.getenv("PINECONE_ENVIRONMENT").lower().endswith(starter_env_suffixes)

    async def aquery(self,
                     queries: List[Query],
                     global_metadata_filter: Optional[dict] = None
                     ) -> List[QueryResult]:
        raise NotImplementedError()

    async def aupsert(self,
                      documents: List[Document],
                      namespace: str = "") -> None:
        raise NotImplementedError()

    async def adelete(self,
                      document_ids: List[str],
                      namespace: str = "") -> None:
        raise NotImplementedError()



from typing import Dict, Any


class ConfigurableMixin:
    _DEFAULT_COMPONENTS: Dict[str, type] = {}

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if not hasattr(cls, '_SUPPORTED_CLASSES'):
            cls._SUPPORTED_CLASSES = {}
        if ConfigurableMixin in cls.__bases__:
            cls.__FACTORY_BASE_CLASS__ = cls
        else:
            cls._SUPPORTED_CLASSES[cls.__name__] = cls

    @classmethod
    def from_config(cls, config: Dict[str, Any]):
        return cls._from_config(config)

    @classmethod
    def _from_config(cls, config: Dict[str, Any]):
        # Make a copy of the config, so we can modify it (e.g. pop fields) without
        # affecting the user's original config.
        config = config.copy()

        # These asserts should be true for all subclasses of ConfigurableMixin
        assert hasattr(cls, '_SUPPORTED_CLASSES')
        assert hasattr(cls, '__FACTORY_BASE_CLASS__')

        # If this is the base class, we expect a 'type' field in the config which will
        # tell us which derived class to load.
        if cls is cls.__FACTORY_BASE_CLASS__:
            derived_class = cls._get_derived_class(config)
            return derived_class.from_config(config)

        # If we got here, this is a derived class. We do not expect a 'type' field
        # in the config since we already know which class to load.
        if "type" in config:
            base_name = cls.__FACTORY_BASE_CLASS__.__name__
            raise ValueError(
                f"{cls.__name__} load error: 'type' field is not allowed in config. "
                f"if you wish to load another {base_name} subclass, use "
                f"`{base_name}.from_config(config)` instead."
            )

        # Load the class's subcomponents (dependencies) recursively
        loaded_components = cls._load_sub_components(config)

        parameters = config.pop("params", {})

        # The config should be empty at this point
        if config:
            allowed_keys = ['type', 'params'] + list(cls._DEFAULT_COMPONENTS.keys())
            raise ValueError(
                f"Unrecognized keys in {cls.__name__} config: {list(config.keys())}. "
                f"The allowed keys are: {allowed_keys}"
            )

        try:
            return cls(**loaded_components, **parameters)
        except TypeError as e:
            raise TypeError(
                f"{cls.__name__} load error: {e}. Please check the config."
            )

    @classmethod
    def _get_derived_class(cls, config):
        if "type" not in config:
            raise ValueError(
                f"{cls.__name__} load error: missing 'type' field in config."
            )
        derived_class_name = config.pop("type")
        if derived_class_name not in cls._SUPPORTED_CLASSES:
            raise ValueError(
                f"{cls.__name__} load error: {derived_class_name} is not supported."
                f" Supported types are: {list(cls._SUPPORTED_CLASSES.keys())}"
            )
        derived_class = cls._SUPPORTED_CLASSES[derived_class_name]
        return derived_class

    @classmethod
    def list_supported_types(cls):
        if cls is not cls.__FACTORY_BASE_CLASS__:
            raise RuntimeError(
                f"{cls.__name__} list_supported_types() should only be called on the "
                f"base class."
            )
        return list(cls._SUPPORTED_CLASSES.keys())

    @classmethod
    def _load_sub_components(cls, config):
        loaded_components = {}
        for component_name, default_class in cls._DEFAULT_COMPONENTS.items():
            component_config = config.pop(component_name, {})
            component_config['type'] = component_config.get(
                'type', default_class.__name__
            )

            # For classes implementing ConfigurableMixin, we need to call
            # `from_config()` on the base class
            assert hasattr(default_class, '__FACTORY_BASE_CLASS__')
            base_class = default_class.__FACTORY_BASE_CLASS__
            component = base_class.from_config(component_config)
            loaded_components[component_name] = component
        return loaded_components



from typing import List

from canopy.chat_engine.query_generator import QueryGenerator
from canopy.models.data_models import Messages, Query, Role


class LastMessageQueryGenerator(QueryGenerator):
    """
        Returns the last message as a query without running any LLMs. This can be
        considered as the most basic query generation. Please use other query generators
        for more accurate results.
    """

    def generate(self,
                 messages: Messages,
                 max_prompt_tokens: int) -> List[Query]:
        """
            max_prompt_token is dismissed since we do not consume any token for
            generating the queries.
        """

        if len(messages) == 0:
            raise ValueError("Passed chat history does not contain any messages. "
                             "Please include at least one message in the history.")

        last_message = messages[-1]

        if last_message.role != Role.USER:
            raise ValueError(f"Expected a UserMessage, got {type(last_message)}.")

        return [Query(text=last_message.content)]

    async def agenerate(self,
                        messages: Messages,
                        max_prompt_tokens: int) -> List[Query]:
        return self.generate(messages, max_prompt_tokens)



import json
import os
import glob
from enum import Enum
from collections.abc import Iterable
from typing import List
from textwrap import dedent

import numpy as np
import pandas as pd

from pydantic import ValidationError

from canopy.models.data_models import Document
from canopy_cli.data_loader.errors import (
    DataLoaderException,
    DocumentsValidationError,
    IDsNotUniqueError)


class NonSchematicFilesTypes(Enum):
    TEXT = "txt"


def format_multiline(msg):
    return dedent(msg).strip()


def _process_metadata(value):
    if pd.isna(value):
        return {}

    if isinstance(value, str):
        try:
            value = json.loads(value)
        except json.JSONDecodeError as e:
            raise DocumentsValidationError(
                f"Metadata must be a valid json string. Error: {e}"
            ) from e

    if not isinstance(value, dict):
        raise DocumentsValidationError("Metadata must be a dict or json string")

    return {k: v.tolist() if isinstance(v, np.ndarray) else v
            for k, v in value.items()
            if isinstance(v, Iterable) or pd.notna(v)}


def _df_to_documents(df: pd.DataFrame, origin_file_path=None) -> List[Document]:
    if not isinstance(df, pd.DataFrame):
        raise ValueError("Dataframe must be a pandas DataFrame")
    if "id" not in df.columns:
        raise DocumentsValidationError("Missing 'id' column")
    if df.id.nunique() != df.shape[0]:
        raise IDsNotUniqueError("IDs must be unique")

    try:
        if "metadata" in df.columns:
            df.loc[:, "metadata"] = df["metadata"].apply(_process_metadata)
        documents = []
        for row in df.itertuples(index=False):
            try:
                documents.append(
                    Document(
                        **{k: v for k, v in row._asdict().items() if not pd.isna(v)}
                    )
                )
            except ValidationError as e:
                raise DataLoaderException(
                    file_name=origin_file_path,
                    row_id=row.id,
                    err=format_multiline(e.errors()[0]["msg"])
                ) from e
    except ValidationError as e:
        raise DocumentsValidationError("Documents failed validation") from e
    except ValueError as e:
        raise DocumentsValidationError(f"Unexpected error in validation: {e}") from e
    return documents


def _load_multiple_txt_files(file_paths: List[str]) -> pd.DataFrame:
    """Load multiple text files into a single dataframe

    Args:
        file_paths: List of file paths to load

    Returns:
        pd.DataFrame: Dataframe with columns `id`, `text` and 'source`
                      Note: metadata will be empty
    """
    if not isinstance(file_paths, list):
        raise ValueError("file_paths must be a list of strings")
    if len(file_paths) == 0:
        raise ValueError("file_paths must not be empty")

    rows = []
    for file_path in file_paths:
        try:
            with open(file_path, "r", encoding='utf-8') as f:
                text = f.read()
                rows.append(
                    {
                        "id": os.path.basename(file_path).replace(".txt", ""),
                        "text": text,
                        "source": file_path
                    }
                )
        except UnicodeDecodeError as e:
            raise DataLoaderException(
                file_name=file_path,
                row_id="*",
                err="File must be UTF-8 encoded"
            ) from e
    df = pd.DataFrame(rows, columns=["id", "text", "source"])
    return df


def _load_single_schematic_file_by_suffix(file_path: str) -> List[Document]:
    try:
        if file_path.endswith(".parquet"):
            df = pd.read_parquet(file_path)
        elif file_path.endswith(".csv"):
            df = pd.read_csv(file_path)
        elif file_path.endswith(".jsonl"):
            df = pd.read_json(file_path, lines=True)
        else:
            raise ValueError(
                "Only [.parquet, .jsonl, .csv, .txt] files are supported"
            )
    except ValueError as e:
        raise e
    except Exception as e:
        raise DataLoaderException(
            file_name=file_path,
            row_id="*",
            err=str(e)
        ) from e
    return _df_to_documents(df, origin_file_path=file_path)


def _load_multiple_non_schematic_files(
    file_paths: List[str],
    type: NonSchematicFilesTypes
) -> List[Document]:
    if not isinstance(file_paths, list):
        raise ValueError("file_paths must be a list of strings")
    if len(file_paths) == 0:
        raise ValueError("file_paths must not be empty")

    if type == NonSchematicFilesTypes.TEXT:
        df = _load_multiple_txt_files(file_paths)
    else:
        raise ValueError(f"Unsupported file type: {type}")

    return _df_to_documents(df)


def load_from_path(path: str) -> List[Document]:
    """
    Load documents from a file or directory

    Args:
        path: Path to file or directory

    Returns:
        List[Document]: List of documents
    """
    if os.path.isdir(path):
        # List all files in directory
        all_files_schematic = []
        all_files_non_schematic_txt = []
        for file in glob.glob(os.path.join(path, "*")):
            if not os.path.isfile(file):
                continue
            if file.endswith(".txt"):
                all_files_non_schematic_txt.append(file)
            elif (file.endswith(".jsonl") or
                  file.endswith(".csv") or
                  file.endswith(".parquet")):
                all_files_schematic.append(file)
        if len(all_files_schematic) + len(all_files_non_schematic_txt) == 0:
            raise ValueError("No files found in directory")

        documents: List[Document] = []
        # Load all schematic files
        for f in all_files_schematic:
            documents.extend(_load_single_schematic_file_by_suffix(f))

        # Load all non-schematic files
        if len(all_files_non_schematic_txt) > 0:
            documents.extend(
                _load_multiple_non_schematic_files(
                    all_files_non_schematic_txt,
                    NonSchematicFilesTypes.TEXT))

    # Load single file
    elif os.path.isfile(path):
        if path.endswith(".txt"):
            documents = _load_multiple_non_schematic_files(
                [path],
                NonSchematicFilesTypes.TEXT)
        else:
            documents = _load_single_schematic_file_by_suffix(path)
    else:
        raise ValueError(f"Could not find file or directory at {path}")
    return documents



import typing as t
from click._compat import get_text_stderr
from click import echo


class IDsNotUniqueError(ValueError):
    pass


class DocumentsValidationError(ValueError):
    pass


class DataLoaderException(Exception):
    """An exception that Click can handle and show to the user."""

    #: The exit code for this exception.
    exit_code = 1

    def __init__(self, file_name: str, row_id: str, err: str) -> None:
        message = f"""
        {file_name}, line {row_id} - {err}
        """
        super().__init__(message)
        self.file_name = file_name
        self.row_id = row_id
        self.err = err

    def format_message(self) -> str:
        message = f"""
        {self.file_name}, line {self.row_id} - {self.err}
        """
        return message

    def __str__(self) -> str:
        return self.format_message()

    def show(self, file: t.Optional[t.IO] = None) -> None:
        if file is None:
            file = get_text_stderr()

        echo("{message}".format(message=self.format_message()), file=file)



import os
import signal
import subprocess
from typing import Dict, Any, Optional, List, Iterable

import click
from prompt_toolkit import prompt
import time

import requests
import yaml
from dotenv import load_dotenv
from tenacity import retry, stop_after_attempt, wait_fixed
from tqdm import tqdm

import pandas as pd
import openai
from openai import APIError as OpenAI_APIError
from urllib.parse import urljoin

from canopy.knowledge_base import KnowledgeBase
from canopy.knowledge_base import connect_to_pinecone
from canopy.knowledge_base.chunker import Chunker
from canopy.chat_engine import ChatEngine
from canopy.models.data_models import Document, UserMessage
from canopy.tokenizer import Tokenizer
from canopy_cli.data_loader import (
    load_from_path,
    IDsNotUniqueError,
    DocumentsValidationError)
from canopy_cli.errors import CLIError

from canopy import __version__

from canopy_server.app import start as start_server, API_VERSION
from .cli_spinner import Spinner
from canopy_server.models.v1.api_models import ChatDebugInfo


load_dotenv()


CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])
DEFAULT_SERVER_URL = f"http://localhost:8000/{API_VERSION}"
spinner = Spinner()


def check_server_health(url: str):
    try:
        res = requests.get(urljoin(url, "/health"))
        res.raise_for_status()
        return res.ok
    except requests.exceptions.ConnectionError:
        msg = f"""
        Canopy server is not running on {url}.
        please run `canopy start`
        """
        raise CLIError(msg)

    except requests.exceptions.HTTPError as e:
        if e.response is not None:
            error = e.response.json().get("detail", None) or e.response.text
        else:
            error = str(e)
        msg = (
            f"Canopy server on {url} is not healthy, failed with error: {error}"
        )
        raise CLIError(msg)


@retry(reraise=True, wait=wait_fixed(5), stop=stop_after_attempt(6))
def wait_for_server(chat_server_url: str):
    check_server_health(chat_server_url)


def validate_pinecone_connection():
    try:
        connect_to_pinecone()
    except RuntimeError as e:
        msg = (
            f"{str(e)}\n"
            "Credentials should be set by the PINECONE_API_KEY and PINECONE_ENVIRONMENT"
            " environment variables.\n"
            "Please visit https://www.pinecone.io/docs/quickstart/ for more details."
        )
        raise CLIError(msg)


def _initialize_tokenizer():
    try:
        Tokenizer.initialize()
    except Exception as e:
        msg = f"Failed to initialize tokenizer. Reason:\n{e}"
        raise CLIError(msg)


def _read_config_file(config_file: Optional[str]) -> Dict[str, Any]:
    if config_file is None:
        return {}

    try:
        with open(config_file, 'r') as f:
            config = yaml.safe_load(f)
    except Exception as e:
        msg = f"Failed to load config file {config_file}. Reason:\n{e}"
        raise CLIError(msg)

    return config


def _load_kb_config(config_file: Optional[str]) -> Dict[str, Any]:
    config = _read_config_file(config_file)
    if not config:
        return {}

    if "knowledge_base" in config:
        kb_config = config.get("knowledge_base", None)
    elif "chat_engine" in config:
        kb_config = config["chat_engine"]\
            .get("context_engine", {})\
            .get("knowledge_base", None)
    else:
        kb_config = None

    if kb_config is None:
        msg = (f"Did not find a `knowledge_base` configuration in {config_file}, "
               "Would you like to use the default configuration?")
        click.confirm(click.style(msg, fg="red"), abort=True)
        kb_config = {}
    return kb_config


def _validate_chat_engine(config_file: Optional[str]):
    config = _read_config_file(config_file)
    Tokenizer.initialize()
    try:
        # If the server itself will fail, we can't except the error, since it's running
        # in a different process. Try to load and run the ChatEngine so we can catch
        # any errors and print a nice message.
        chat_engine = ChatEngine.from_config(config.get("chat_engine", {}))
        chat_engine.max_generated_tokens = 5
        chat_engine.context_engine.knowledge_base.connect()
        chat_engine.chat(
            [UserMessage(content="This is a health check. Are you alive? Be concise")]
        )
    except Exception as e:
        msg = f"Failed to initialize Canopy server. Reason:\n{e}"
        if config_file:
            msg += f"\nPlease check the configuration file {config_file}"
        raise CLIError(msg)
    finally:
        Tokenizer.clear()


class CanopyCommandGroup(click.Group):
    """
    A custom click Group that lets us control the order of commands in the help menu.
    """
    def __init__(self, name=None, commands=None, **attrs):
        super().__init__(name, commands, **attrs)
        self._commands_order = {
            "new": 0,
            "upsert": 1,
            "start": 2,
            "chat": 3,
            "health": 4,
            "stop": 5,
            "api-docs": 6,

        }

    def list_commands(self, ctx):
        return sorted(self.commands, key=lambda x: self._commands_order.get(x, 1000))


@click.group(invoke_without_command=True, context_settings=CONTEXT_SETTINGS,
             cls=CanopyCommandGroup)
@click.version_option(__version__, "-v", "--version", prog_name="Canopy")
@click.pass_context
def cli(ctx):
    """
    \b
    CLI for Pinecone Canopy. Actively developed by Pinecone.
    To use the CLI, you need to have a Pinecone account.
    Visit https://www.pinecone.io/ to sign up for free.
    """
    if ctx.invoked_subcommand is None:
        click.echo(ctx.get_help())


@cli.command(help="Check if the Canopy server is running and healthy.")
@click.option("--url", default=DEFAULT_SERVER_URL,
              help=("Canopy's server url. "
                    f"Defaults to {DEFAULT_SERVER_URL}"))
def health(url):
    check_server_health(url)
    click.echo(click.style("Canopy server is healthy!", fg="green"))
    return


@cli.command(
    help=(
        """
        \b
        Create a new Pinecone index that will be used by Canopy.

        A Canopy service cannot be started without a Pinecone index that is configured
        to work with Canopy. This command creates a new Pinecone index and configures
        it in the right schema.

        If the embedding vectors' dimension is not explicitly configured in
        the config file, the embedding model will be tapped with a single token to
        infer the dimensionality of the embedding space.
        """  # noqa: E501
    )
)
@click.argument("index-name", nargs=1, envvar="INDEX_NAME", type=str, required=True)
@click.option("--config", "-c", default=None, envvar="CANOPY_CONFIG_FILE",
              help="Path to a canopy config file. Can also be set by the "
                   "`CANOPY_CONFIG_FILE` envrionment variable. Otherwise, the built-in"
                   "defualt configuration will be used.")
def new(index_name: str, config: Optional[str]):
    _initialize_tokenizer()
    kb_config = _load_kb_config(config)
    kb = KnowledgeBase.from_config(kb_config, index_name=index_name)
    click.echo("Canopy is going to create a new index: ", nl=False)
    click.echo(click.style(f"{kb.index_name}", fg="green"))
    click.confirm(click.style("Do you want to continue?", fg="red"), abort=True)
    with spinner:
        try:
            kb.create_canopy_index()
        # TODO: kb should throw a specific exception for failure
        except Exception as e:
            already_exists_str = f"Index {kb.index_name} already exists"
            if isinstance(e, RuntimeError) and already_exists_str in str(e):
                msg = (f"{already_exists_str}, please use a different name."
                       f"If you wish to delete the index, log in to Pinecone's "
                       f"Console: https://app.pinecone.io/")
            else:
                msg = f"Failed to create a new index. Reason:\n{e}"
            raise CLIError(msg)
    click.echo(click.style("Success!", fg="green"))
    os.environ["INDEX_NAME"] = index_name


def _batch_documents_by_chunks(chunker: Chunker,
                               documents: List[Document],
                               batch_size: int = 400) -> Iterable[List[Document]]:
    """
    Note: this is a temporary solution until we improve the upsert pipeline.
          using the chunker directly is not recommended, especially since the knowledge base also going to use it internally on the same documents.
    """  # noqa: E501
    num_chunks_in_batch = 0
    batch: List[Document] = []
    for doc in documents:
        cur_num_chunks = len(chunker.chunk_single_document(doc))
        if num_chunks_in_batch + cur_num_chunks >= batch_size:
            yield batch
            batch = []
            num_chunks_in_batch = 0
        num_chunks_in_batch += cur_num_chunks
        batch.append(doc)
    if batch:
        yield batch


@cli.command(
    help=(
        """
        \b
        Upload local data files to the Canopy service.

        Load all the documents from a data file or a directory containing multiple data
        files. The allowed formats are .jsonl, .parquet, .csv, and .txt.
        """  # noqa: E501
    )
)
@click.argument("data-path", type=click.Path(exists=True))
@click.option(
    "--index-name",
    default=os.environ.get("INDEX_NAME"),
    help="The name of the index to upload the data to. "
         "Inferred from INDEX_NAME env var if not provided."
)
@click.option("--allow-failures/--dont-allow-failures", default=False,
              help="On default, the upsert process will stop if any document fails to "
                   "be uploaded. "
                   "When set to True, the upsert process will continue on failure, as "
                   "long as less than 10% of the documents have failed to be uploaded.")
@click.option("--config", "-c", default=None, envvar="CANOPY_CONFIG_FILE",
              help="Path to a canopy config file. Can also be set by the "
                   "`CANOPY_CONFIG_FILE` envrionment variable. Otherwise, the built-in"
                   "defualt configuration will be used.")
def upsert(index_name: str,
           data_path: str,
           allow_failures: bool,
           config: Optional[str]):
    if index_name is None:
        msg = (
            "No index name provided. Please set --index-name or INDEX_NAME environment "
            "variable."
        )
        raise CLIError(msg)

    validate_pinecone_connection()

    _initialize_tokenizer()

    kb_config = _load_kb_config(config)
    try:
        kb = KnowledgeBase.from_config(kb_config, index_name=index_name)
    except Exception as e:
        raise CLIError(str(e))

    try:
        kb.connect()
    except RuntimeError as e:
        # TODO: kb should throw a specific exception for each case
        msg = str(e)
        raise CLIError(msg)

    click.echo("Canopy is going to upsert data from ", nl=False)
    click.echo(click.style(f'{data_path}', fg='yellow'), nl=False)
    click.echo(" to index: ")
    click.echo(click.style(f'{kb.index_name} \n', fg='green'))
    with spinner:
        try:
            data = load_from_path(data_path)
        except IDsNotUniqueError:
            msg = (
                "The data contains duplicate IDs. Please make sure that each document"
                " has a unique ID; otherwise, documents with the same ID will overwrite"
                " each other."
            )
            raise CLIError(msg)
        except DocumentsValidationError:
            msg = (
                f"One or more rows have failed data validation. The rows in the"
                f"data file should be in the schema: {Document.__annotations__}."
            )
            raise CLIError(msg)
        except Exception:
            msg = (
                f"A unexpected error while loading the data from files in {data_path}. "
                "Please make sure the data is in valid `jsonl`, `parquet`, or `csv`"
                " format, or plaintext `.txt` files."
            )
            raise CLIError(msg)
        pd.options.display.max_colwidth = 20
    click.echo(pd.DataFrame([doc.dict(exclude_none=True) for doc in data[:5]]))
    click.echo(click.style(f"\nTotal records: {len(data)}"))
    click.confirm(click.style("\nDoes this data look right?", fg="red"),
                  abort=True)

    pbar = tqdm(total=len(data), desc="Upserting documents")
    failed_docs: List[str] = []
    first_error: Optional[str] = None
    for batch in _batch_documents_by_chunks(kb._chunker, data,
                                            batch_size=kb._encoder.batch_size):
        try:
            kb.upsert(batch)
        except Exception as e:
            if allow_failures and len(failed_docs) < len(data) // 10:
                failed_docs.extend([_.id for _ in batch])
                if first_error is None:
                    first_error = str(e)
            else:
                msg = (
                    f"Failed to upsert data to index {kb.index_name}. "
                    f"Underlying error: {e}\n"
                    f"You can allow partial failures by setting --allow-failures. "
                )
                raise CLIError(msg)

        pbar.update(len(batch))
    pbar.close()

    if failed_docs:
        msg = (
            f"Failed to upsert the following documents to index {kb.index_name}: "
            f"{failed_docs}. The first encountered error was: {first_error}"
        )
        raise CLIError(msg)

    click.echo(click.style("Success!", fg="green"))


def _chat(
    speaker,
    speaker_color,
    model,
    history,
    message,
    openai_api_key=None,
    api_base=None,
    stream=True,
    print_debug_info=False,
):
    if openai_api_key is None:
        openai_api_key = os.environ.get("OPENAI_API_KEY")
    if openai_api_key is None and api_base is None:
        raise CLIError(
            "No OpenAI API key provided. When using the `--no-rag` flag "
            "You will need to have a valid OpenAI API key. "
            "Please set the OPENAI_API_KEY environment "
            "variable."
        )
    output = ""
    history += [{"role": "user", "content": message}]
    client = openai.OpenAI(base_url=api_base, api_key=openai_api_key)

    start = time.time()
    try:
        openai_response = client.chat.completions.create(
            model=model, messages=history, stream=stream
        )
    except (Exception, OpenAI_APIError) as e:
        err = e.http_body if isinstance(e, OpenAI_APIError) else str(e)
        msg = f"Oops... something went wrong. The error I got is: {err}"
        raise CLIError(msg)
    end = time.time()
    duration_in_sec = end - start
    click.echo(click.style(f"\n {speaker}:\n", fg=speaker_color))
    if stream:
        for chunk in openai_response:
            openai_response_id = chunk.id
            internal_model = chunk.model
            text = chunk.choices[0].delta.content or ""
            output += text
            click.echo(text, nl=False)
        click.echo()
        debug_info = ChatDebugInfo(
            id=openai_response_id,
            internal_model=internal_model,
            duration_in_sec=round(duration_in_sec, 2),
        )
    else:
        internal_model = openai_response.model
        text = openai_response.choices[0].message.content or ""
        output = text
        click.echo(text, nl=False)
        debug_info = ChatDebugInfo(
            id=openai_response.id,
            internal_model=internal_model,
            duration_in_sec=duration_in_sec,
            prompt_tokens=openai_response.usage.prompt_tokens,
            generated_tokens=openai_response.usage.completion_tokens,
        )
    if print_debug_info:
        click.echo()
        click.echo(
            click.style(f"{debug_info.to_text()}", fg="bright_black", italic=True)
        )
    history += [{"role": "assistant", "content": output}]
    return debug_info


@cli.command(
    help=(
        """
        Debugging tool for chatting with the Canopy RAG service.

        Run an interactive chat with the Canopy RAG service, for debugging and demo
        purposes. A prompt is provided for the user to enter a message, and the
        RAG-infused ChatBot will respond. You can continue the conversation by entering
        more messages. Hit Ctrl+C to exit.

        To compare RAG-infused ChatBot with the original LLM, run with the `--no-rag`
        flag, which would display both models' responses side by side.
        """

    )
)
@click.option("--stream/--no-stream", default=True,
              help="Stream the response from the RAG chatbot word by word.")
@click.option("--debug/--no-debug", default=False,
              help="Print additional debugging information.")
@click.option("--rag/--no-rag", default=True,
              help="Compare RAG-infused Chatbot with vanilla LLM.",)
@click.option("--chat-server-url", default=DEFAULT_SERVER_URL,
              help=("URL of the Canopy server to use."
                    f" Defaults to {DEFAULT_SERVER_URL}"))
def chat(chat_server_url, rag, debug, stream):
    check_server_health(chat_server_url)
    note_msg = (
        "🚨 Note 🚨\n"
        "Chat is a debugging tool; it is not meant to be used for production!"
    )
    for c in note_msg:
        click.echo(click.style(c, fg="red"), nl=False)
        if (stream):
            time.sleep(0.01)
    click.echo()
    note_white_message = (
        "This method should be used by developers to test the RAG data and model "
        "during development. "
        "When you are ready to deploy, run the Canopy server as a REST API "
        "backend for your chatbot UI. \n\n"
        "Let's Chat!"
    )
    for c in note_white_message:
        click.echo(click.style(c, fg="white"), nl=False)
        if (stream):
            time.sleep(0.01)
    click.echo()

    history_with_pinecone = []
    history_without_pinecone = []

    while True:
        click.echo(click.style("\nUser message: ", fg="bright_blue"), nl=False)
        click.echo(
            click.style("([Esc] followed by [Enter] to accept input)\n",
                        italic=True,
                        fg="bright_black"),
            nl=True
        )
        message = prompt("  ", multiline=True, )

        if not message:
            click.echo(click.style("Please enter a message", fg="red"))
            continue

        if message == "exit":
            click.echo(click.style("Bye!", fg="red"))
            break

        if message.isspace() or message == "":
            click.echo(click.style("Please enter a message", fg="red"))
            continue

        dubug_info = _chat(
            speaker="With Context (RAG)",
            speaker_color="green",
            model="",
            history=history_with_pinecone,
            message=message,
            stream=stream,
            openai_api_key="canopy",
            api_base=chat_server_url,
            print_debug_info=debug,
        )

        if not rag:
            _ = _chat(
                speaker="Without Context (No RAG)",
                speaker_color="yellow",
                model=dubug_info.internal_model,
                history=history_without_pinecone,
                message=message,
                stream=stream,
                print_debug_info=debug,
            )

        click.echo(click.style("\n.", fg="bright_black"))
        click.echo(
            click.style(
                f"| {len(history_with_pinecone) // 2}", fg="bright_black", bold=True
            ),
            nl=True,
        )
        click.echo(click.style("˙▔▔▔", fg="bright_black", bold=True), nl=False)
        click.echo(click.style("˙", fg="bright_black", bold=True))


@cli.command(
    help=(
        """
        \b
        Start the Canopy server.

        This command launches a Uvicorn server to serve the Canopy API.

        If you would like to try out the chatbot, run `canopy chat` in a separate
        terminal window.
        """
    )
)
@click.option("--stream/--no-stream", default=True,
              help="Stream the response from the RAG chatbot word by word.")
@click.option("--host", default="0.0.0.0",
              help="Hostname or address to bind the server to. Defaults to 0.0.0.0")
@click.option("--port", default=8000,
              help="TCP port to bind the server to. Defaults to 8000")
@click.option("--reload/--no-reload", default=False,
              help="Set the server to reload on code changes. Defaults to False")
@click.option("--config", "-c", default=None, envvar="CANOPY_CONFIG_FILE",
              help="Path to a canopy config file. Can also be set by the "
                   "`CANOPY_CONFIG_FILE` envrionment variable. Otherwise, the built-in"
                   "defualt configuration will be used.")
@click.option("--index-name", default=None,
              help="Index name, if not provided already as an environment variable.")
def start(host: str, port: str, reload: bool, stream: bool,
          config: Optional[str], index_name: Optional[str]):
    validate_pinecone_connection()
    _validate_chat_engine(config)

    note_msg = (
        "🚨 Note 🚨\n"
        "For debugging only. To run the Canopy server in production, "
    )
    msg_suffix = (
        "run the command:"
        "\n"
        "gunicorn canopy_server.app:app --worker-class uvicorn.workers.UvicornWorker "
        f"--bind {host}:{port} --workers <num_workers>"
    ) if os.name != "nt" else (
        # TODO: Replace with proper instructions once we have a Dockerfile
        "please use Docker with a Gunicorn server."
    )
    for c in note_msg + msg_suffix:
        click.echo(click.style(c, fg="red"), nl=False)
        if (stream):
            time.sleep(0.01)
    click.echo()

    if index_name:
        env_index_name = os.getenv("INDEX_NAME")
        if env_index_name and index_name != env_index_name:
            raise CLIError(
                f"Index name provided via --index-name '{index_name}' does not match "
                f"the index name provided via the INDEX_NAME environment variable "
                f"'{env_index_name}'"
            )
        os.environ["INDEX_NAME"] = index_name

    click.echo(f"Starting Canopy server on {host}:{port}")
    start_server(host, port=port, reload=reload, config_file=config)


@cli.command(
    help=(
        """
        \b
        Stop the Canopy server.

        This command sends a shutdown request to the Canopy server.
        """
    )
)
@click.option("url", "--url", default=DEFAULT_SERVER_URL,
              help=("URL of the Canopy server to use. "
                    f"Defaults to {DEFAULT_SERVER_URL}"))
def stop(url):
    if os.name != "nt":
        # Check if the server was started using Gunicorn
        res = subprocess.run(["pgrep", "-f", "gunicorn canopy_server.app:app"],
                             capture_output=True)
        output = res.stdout.decode("utf-8").split()

        # If Gunicorn was used, kill all Gunicorn processes
        if output:
            msg = ("It seems that Canopy server was launched using Gunicorn.\n"
                   "Do you want to kill all Gunicorn processes?")
            click.confirm(click.style(msg, fg="red"), abort=True)
            try:
                subprocess.run(["pkill", "-f", "gunicorn canopy_server.app:app"],
                               check=True)
            except subprocess.CalledProcessError:
                try:
                    [os.kill(int(pid), signal.SIGINT) for pid in output]
                except OSError:
                    msg = (
                        "Could not kill Gunicorn processes. Please kill them manually."
                        f"Found process ids: {output}"
                    )
                    raise CLIError(msg)

    try:
        res = requests.get(urljoin(url, "/shutdown"))
        res.raise_for_status()
        return res.ok
    except requests.exceptions.ConnectionError:
        msg = f"""
        Could not find Canopy server on {url}.
        """
        raise CLIError(msg)


@cli.command(
    help=(
        """
        \b
        Open the Canopy server docs.
        """
    )
)
@click.option("--url", default="http://localhost:8000",
              help="Canopy's server url. Defaults to http://localhost:8000")
def api_docs(url):
    import webbrowser

    generated_docs = False
    try:
        check_server_health(url)
    except CLIError:
        msg = ("Canopy server is not running. Would you like to generate the docs "
               "to a local HTML file?")
        click.confirm(click.style(msg, fg="red"), abort=True)
        generated_docs = True

    if generated_docs:
        import json
        from canopy_server._redocs_template import HTML_TEMPLATE
        from canopy_server.app import app, _init_routes
        # generate docs
        _init_routes(app)
        filename = "canopy-api-docs.html"
        msg = f"Generating docs to {filename}"
        click.echo(click.style(msg, fg="green"))
        with open(filename, "w") as fd:
            print(HTML_TEMPLATE % json.dumps(app.openapi()), file=fd)
        webbrowser.open('file://' + os.path.realpath(filename))
    else:
        webbrowser.open(urljoin(url, "redoc"))


if __name__ == "__main__":
    cli()



from typing import List, Optional

from canopy.chat_engine.models import HistoryPruningMethod
from canopy.chat_engine.prompt_builder import PromptBuilder
from canopy.chat_engine.query_generator import QueryGenerator
from canopy.llm import BaseLLM, OpenAILLM
from canopy.llm.models import (Function, FunctionParameters,
                               FunctionArrayProperty)
from canopy.models.data_models import Messages, Query

DEFAULT_SYSTEM_PROMPT = """Your task is to formulate search queries for a search engine, to assist in responding to the user's question.
You should break down complex questions into sub-queries if needed.
"""  # noqa: E501

DEFAULT_FUNCTION_DESCRIPTION = """Query search engine for relevant information"""


class FunctionCallingQueryGenerator(QueryGenerator):

    _DEFAULT_COMPONENTS = {
        "llm": OpenAILLM,
    }

    def __init__(self,
                 *,
                 llm: Optional[BaseLLM] = None,
                 prompt: Optional[str] = None,
                 function_description: Optional[str] = None):
        self._llm = llm or self._DEFAULT_COMPONENTS["llm"]()
        self._system_prompt = prompt or DEFAULT_SYSTEM_PROMPT
        self._function_description = \
            function_description or DEFAULT_FUNCTION_DESCRIPTION
        self._prompt_builder = PromptBuilder(HistoryPruningMethod.RAISE, 1)

    def generate(self,
                 messages: Messages,
                 max_prompt_tokens: int) -> List[Query]:
        messages = self._prompt_builder.build(system_prompt=self._system_prompt,
                                              history=messages,
                                              max_tokens=max_prompt_tokens)
        arguments = self._llm.enforced_function_call(messages,
                                                     function=self._function)

        return [Query(text=q)
                for q in arguments["queries"]]

    async def agenerate(self,
                        messages: Messages,
                        max_prompt_tokens: int) -> List[Query]:
        raise NotImplementedError

    @property
    def _function(self) -> Function:
        return Function(
            name="query_knowledgebase",
            description=self._function_description,
            parameters=FunctionParameters(
                required_properties=[
                    FunctionArrayProperty(
                        name="queries",
                        items_type="string",
                        description='List of queries to send to the search engine.',
                    ),
                ]
            ),
        )



from typing import List

from openai import OpenAIError, RateLimitError, APIConnectionError, AuthenticationError
from pinecone_text.dense.openai_encoder import OpenAIEncoder
from canopy.knowledge_base.models import KBDocChunk, KBEncodedDocChunk, KBQuery
from canopy.knowledge_base.record_encoder.dense import DenseRecordEncoder
from canopy.models.data_models import Query


def _format_openai_error(e):
    try:
        return e.response.json()['error']['message']
    except Exception:
        return str(e)


class OpenAIRecordEncoder(DenseRecordEncoder):
    """
    OpenAIRecordEncoder is a type of DenseRecordEncoder that uses the OpenAI `embeddings` API.
    The implementation uses the `OpenAIEncoder` class from the `pinecone-text` library.
    For more information about see: https://github.com/pinecone-io/pinecone-text

    """  # noqa: E501

    def __init__(self,
                 *,
                 model_name: str = "text-embedding-ada-002",
                 batch_size: int = 400,
                 **kwargs):
        """
        Initialize the OpenAIRecordEncoder

        Args:
            model_name: The name of the OpenAI embeddings model to use for encoding. See https://platform.openai.com/docs/models/embeddings
            batch_size: The number of documents or queries to encode at once.
                        Defaults to 400.
            **kwargs: Additional arguments to pass to the underlying `pinecone-text. OpenAIEncoder`.
        """  # noqa: E501
        try:
            encoder = OpenAIEncoder(model_name, **kwargs)
        except OpenAIError as e:
            raise RuntimeError(
                "Failed to connect to OpenAI, please make sure that the OPENAI_API_KEY "
                "environment variable is set correctly.\n"
                f"Error: {_format_openai_error(e)}"
            ) from e
        super().__init__(dense_encoder=encoder, batch_size=batch_size)

    async def _aencode_documents_batch(self,
                                       documents: List[KBDocChunk]
                                       ) -> List[KBEncodedDocChunk]:
        raise NotImplementedError

    async def _aencode_queries_batch(self, queries: List[Query]) -> List[KBQuery]:
        raise NotImplementedError

    def _format_error(self, err):
        if isinstance(err, RateLimitError):
            return (f"Your OpenAI account seem to have reached the rate limit. "
                    f"Details: {_format_openai_error(err)}")
        elif isinstance(err, (AuthenticationError, APIConnectionError)):
            return (f"Failed to connect to OpenAI, please make sure that the "
                    f"OPENAI_API_KEY environment variable is set correctly. "
                    f"Details: {_format_openai_error(err)}")
        else:
            return _format_openai_error(err)



from typing import List, Optional, Type

from .openai import OpenAITokenizer
from .base import BaseTokenizer
from ..models.data_models import Messages


class Tokenizer:

    """
    Singleton class for tokenization.
    The singleton behavior unify tokenization across the system.

    Usage:

    To initialize the tokenizer, call Tokenizer.initialize(tokenizer_class, *args, **kwargs)
    >>> from canopy.tokenizer import Tokenizer
    >>> Tokenizer.initialize()

    Then, you can instantiate a tokenizer instance by calling Tokenizer() from anywhere in the code and use it:
    >>> tokenizer = Tokenizer()
    >>> tokenizer.tokenize("Hello world!")
    ['Hello', 'world', '!']
    >>> tokenizer.detokenize(['Hello', 'world', '!'])
    'Hello world!'
    """  # noqa: E501

    _instance = None
    _tokenizer_instance: Optional[BaseTokenizer] = None
    _initialized = False

    DEFAULT_TOKENIZER_CLASS = OpenAITokenizer

    def __new__(cls):
        if not cls._initialized:
            raise ValueError("Tokenizer must be initialized using "
                             "Tokenizer.initialize(tokenizer_class, *args, **kwargs)")
        if not cls._instance:
            cls._instance = super(Tokenizer, cls).__new__(cls)
        return cls._instance

    @classmethod
    def initialize(cls,
                   tokenizer_class: Type[BaseTokenizer] = DEFAULT_TOKENIZER_CLASS,
                   **kwargs):
        """
        Initialize the tokenizer singleton.

        Args:
            tokenizer_class: The tokenizer class to use. Must be a subclass of BaseTokenizer. Defaults to OpenAITokenizer.
            **kwargs: Keyword arguments to pass to the underlying `Tokenizer` class constructor.

        Examples:
            Initialize the tokenizer with the default tokenizer class:

            >>> from canopy.tokenizer import Tokenizer
            >>> Tokenizer.initialize()

            Initialize the tokenizer with a custom tokenizer class:

            >>> from canopy.tokenizer import Tokenizer
            >>> from canopy.tokenizer.base import BaseTokenizer
            >>> class MyTokenizer(BaseTokenizer):
            ...     def tokenize(self, text: str) -> List[str]:
            ...         return text.split()
            ...     def detokenize(self, tokens: List[str]) -> str:
            ...         return " ".join(tokens)
            ...     def messages_token_count(self, messages) -> int:
            ...         return sum([self.token_count(message) + 3 for message in messages])
            >>> Tokenizer.initialize(MyTokenizer)

            Then, you can instantiate a tokenizer instance by calling Tokenizer() from anywhere in the code:

            >>> from canopy.tokenizer import Tokenizer
            >>> tokenizer = Tokenizer()
        """  # noqa: E501
        if not issubclass(tokenizer_class, BaseTokenizer):
            raise ValueError("Invalid tokenizer class provided")
        if issubclass(tokenizer_class, Tokenizer):
            raise ValueError("Tokenizer singleton cannot be passed as tokenizer_class")
        cls._tokenizer_instance = tokenizer_class(**kwargs)
        cls._initialized = True

    @classmethod
    def clear(cls):
        """
        Clear the tokenizer singleton.
        """
        cls._instance = None
        cls._tokenizer_instance = None
        cls._initialized = False

    @classmethod
    def initialize_from_config(cls, config: dict):
        """
        Initialize the tokenizer singleton from a config dictionary.
        Used by the config module to initialize the tokenizer from a config file.

        Args:
            config: A dictionary containing the tokenizer configuration. If not provided, the OpenAITokenizer will be used.

        Usage:
            >>> from canopy.tokenizer import Tokenizer
            >>> config = {
            ...     "type": "OpenAITokenizer",
            ...     "model_name": "gpt2"
            ... }
            >>> Tokenizer.initialize_from_config(config)
        """  # noqa: E501
        if cls._initialized:
            raise ValueError("Tokenizer has already been initialized")
        config["type"] = config.get("type", cls.DEFAULT_TOKENIZER_CLASS.__name__)
        cls._tokenizer_instance = BaseTokenizer.from_config(config)
        cls._initialized = True

    def tokenize(self, text: str) -> List[str]:
        """
        Splits a text into tokens.

        Args:
            text: The text to tokenize as a string.

        Returns:
            A list of tokens.
        """
        return self._tokenizer_instance.tokenize(text)  # type: ignore[union-attr]

    def detokenize(self, tokens: List[str]) -> str:
        """
        Joins a list of tokens into a text.

        Args:
            tokens: The tokens to join as a list of strings. Consider using tokenize() first.

        Returns:
            The joined text as a string.
        """  # noqa: E501
        return self._tokenizer_instance.detokenize(tokens)   # type: ignore[union-attr]

    def token_count(self, text: str) -> int:
        """
        Counts the number of tokens in a text.

        Args:
            text: The text to count as a string.

        Returns:
            The number of tokens in the text.
        """
        return self._tokenizer_instance.token_count(text)   # type: ignore[union-attr]

    def messages_token_count(self, messages: Messages) -> int:
        """
        Counts the number of tokens in a Messages object.
        Behind the scenes, for each LLM provider there might be a different overhead for each message in the prompt,
        which is not necessarily the same as the number of tokens in the message text.
        This method takes care of that overhead and returns the total number of tokens in the prompt, as counted by the LLM provider.

        Args:
            messages: The Messages object to count.

        Returns:
            The number of tokens in the Messages object.
        """  # noqa: E501
        return self._tokenizer_instance.messages_token_count(messages)   # type: ignore[union-attr] # noqa: E501



from abc import ABC, abstractmethod

from canopy.chat_engine.exceptions import InvalidRequestError
from canopy.chat_engine.history_pruner import (RaisingHistoryPruner,
                                               RecentHistoryPruner, )
from canopy.chat_engine.history_pruner.base import HistoryPruner
from canopy.chat_engine.models import HistoryPruningMethod
from canopy.tokenizer import Tokenizer
from canopy.models.data_models import Messages, Role, MessageBase


class BasePromptBuilder(ABC):

    def __init__(self,
                 history_pruning: HistoryPruningMethod,
                 min_history_messages: int
                 ):
        self._tokenizer = Tokenizer()
        self._history_pruner: HistoryPruner
        if history_pruning == HistoryPruningMethod.RAISE:
            self._history_pruner = RaisingHistoryPruner(min_history_messages)
        elif history_pruning == HistoryPruningMethod.RECENT:
            self._history_pruner = RecentHistoryPruner(min_history_messages)
        else:
            raise ValueError(f"Unknown history pruning method "
                             f"{history_pruning}.")

    @abstractmethod
    def build(self,
              system_prompt: str,
              history: Messages,
              max_tokens: int
              ) -> Messages:
        pass

    @abstractmethod
    async def abuild(self,
                     messages: Messages,
                     max_tokens: int) -> Messages:
        pass

    def _count_tokens(self, messages: Messages) -> int:
        return sum([len(self._tokenizer.tokenize(message.json()))
                    for message in messages])


class PromptBuilder(BasePromptBuilder):

    def build(self,
              system_prompt: str,
              history: Messages,
              max_tokens: int
              ) -> Messages:
        system_massage = [MessageBase(role=Role.SYSTEM,
                                      content=system_prompt)]
        prompt_tokens = self._tokenizer.messages_token_count(system_massage)
        if prompt_tokens > max_tokens:
            raise InvalidRequestError(
                f'System message tokens {prompt_tokens} exceed max tokens {max_tokens}'
            )

        max_history_tokens = max_tokens - prompt_tokens
        pruned_history, num_tokens = self._history_pruner.build(history,
                                                                max_history_tokens)

        return system_massage + pruned_history

    async def abuild(self,
                     messages: Messages,
                     max_tokens: int
                     ) -> Messages:
        raise NotImplementedError()



from abc import ABC, abstractmethod
from typing import List

from canopy.knowledge_base.models import QueryResult
from canopy.models.data_models import Context
from canopy.utils.config import ConfigurableMixin


class ContextBuilder(ABC, ConfigurableMixin):
    """
    BaseContextBuilder is an abstract class that defines the interface for a context
    builder.
    """

    @abstractmethod
    def build(self,
              query_results: List[QueryResult],
              max_context_tokens: int, ) -> Context:
        pass

    @abstractmethod
    async def abuild(self,
                     query_results: List[QueryResult],
                     max_context_tokens: int, ) -> Context:
        pass



HTML_TEMPLATE = """<!DOCTYPE html>
<html>
<head>
    <meta http-equiv="content-type" content="text/html; charset=UTF-8">
    <title>Canopy API Spec</title>
    <meta charset="utf-8">
    <meta name="viewport" content="width=device-width, initial-scale=1">
    <link rel="shortcut icon" href="https://polybit-apps.s3.amazonaws.com/stdlib/users/pinecone/profile/image.png">
    <style>
        body {
            margin: 0;
            padding: 0;
        }
    </style>
    <style data-styled="" data-styled-version="4.4.1"></style>
</head>
<body>
    <div id="redoc-container"></div>
    <title>Redoc</title>
    <script src="https://cdn.jsdelivr.net/npm/redoc/bundles/redoc.standalone.js"> </script>
    <script>
        var spec = %s;
        Redoc.init(spec, {}, document.getElementById("redoc-container"));
    </script>
</body>
</html>
"""  # noqa: E501



from copy import deepcopy
from typing import Union, Iterable, Optional, Any, Dict, List, cast

import jsonschema
import openai
import json

from openai.types.chat import ChatCompletionToolParam
from tenacity import (
    retry,
    stop_after_attempt,
    retry_if_exception_type,
)
from canopy.llm import BaseLLM
from canopy.llm.models import Function
from canopy.models.api_models import ChatResponse, StreamingChatChunk
from canopy.models.data_models import Messages, Query


def _format_openai_error(e):
    try:
        return e.response.json()['error']['message']
    except Exception:
        return str(e)


class OpenAILLM(BaseLLM):
    """
    OpenAI LLM wrapper built on top of the OpenAI Python client.

    Note: OpenAI requires a valid API key to use this class.
          You can set the "OPENAI_API_KEY" environment variable to your API key.
          Or you can directly set it as follows:
          >>> import openai
          >>> openai.api_key = "YOUR_API_KEY"
    """
    def __init__(self,
                 model_name: str = "gpt-3.5-turbo",
                 *,
                 api_key: Optional[str] = None,
                 organization: Optional[str] = None,
                 base_url: Optional[str] = None,
                 **kwargs: Any,
                 ):
        """
        Initialize the OpenAI LLM.

        Args:
            model_name: The name of the model to use. See https://platform.openai.com/docs/models
            api_key: Your OpenAI API key. Defaults to None (uses the "OPENAI_API_KEY" environment variable).
            organization: Your OpenAI organization. Defaults to None (uses the "OPENAI_ORG" environment variable if set, otherwise uses the "default" organization).
            base_url: The base URL to use for the OpenAI API. Defaults to None (uses the default OpenAI API URL).
            **kwargs: Generation default parameters to use for each request. See https://platform.openai.com/docs/api-reference/chat/create
                    For example, you can set the temperature, top_p etc
                    These params can be overridden by passing a `model_params` argument to the `chat_completion` or `enforced_function_call` methods.
        """  # noqa: E501
        super().__init__(model_name)
        try:
            self._client = openai.OpenAI(api_key=api_key,
                                         organization=organization,
                                         base_url=base_url)
        except openai.OpenAIError as e:
            raise RuntimeError(
                "Failed to connect to OpenAI, please make sure that the OPENAI_API_KEY "
                "environment variable is set correctly.\n"
                f"Error: {_format_openai_error(e)}"
            )

        self.default_model_params = kwargs

    @property
    def available_models(self):
        return [k.id for k in self._client.models.list()]

    def chat_completion(self,
                        messages: Messages,
                        *,
                        stream: bool = False,
                        max_tokens: Optional[int] = None,
                        model_params: Optional[dict] = None,
                        ) -> Union[ChatResponse, Iterable[StreamingChatChunk]]:
        """
        Chat completion using the OpenAI API.

        Note: this function is wrapped in a retry decorator to handle transient errors.

        Args:
            messages: Messages (chat history) to send to the model.
            stream: Whether to stream the response or not.
            max_tokens: Maximum number of tokens to generate. Defaults to None (generates until stop sequence or until hitting max context size).
            model_params: Model parameters to use for this request. Defaults to None (uses the default model parameters).
                          Dictonary of parametrs to override the default model parameters if set on initialization.
                          For example, you can pass: {"temperature": 0.9, "top_p": 1.0} to override the default temperature and top_p.
                          see: https://platform.openai.com/docs/api-reference/chat/create
        Returns:
            ChatResponse or StreamingChatChunk

        Usage:
            >>> from canopy.llm import OpenAILLM
            >>> from canopy.models.data_models import UserMessage
            >>> llm = OpenAILLM()
            >>> messages = [UserMessage(content="Hello! How are you?")]
            >>> result = llm.chat_completion(messages)
            >>> print(result.choices[0].message.content)
            "I'm good, how are you?"
        """  # noqa: E501

        model_params_dict: Dict[str, Any] = deepcopy(self.default_model_params)
        model_params_dict.update(
            model_params or {}
        )

        messages = [m.dict() for m in messages]
        try:
            response = self._client.chat.completions.create(model=self.model_name,
                                                            messages=messages,
                                                            stream=stream,
                                                            max_tokens=max_tokens,
                                                            **model_params_dict)
        except openai.OpenAIError as e:
            provider_name = self.__class__.__name__.replace("LLM", "")
            raise RuntimeError(
                f"Failed to use {provider_name}'s {self.model_name} model for chat "
                f"completion.\n"
                f"Error: {_format_openai_error(e)}"
            )

        def streaming_iterator(response):
            for chunk in response:
                yield StreamingChatChunk.parse_obj(chunk)

        if stream:
            return streaming_iterator(response)

        return ChatResponse.parse_obj(response)

    @retry(
        reraise=True,
        stop=stop_after_attempt(3),
        retry=retry_if_exception_type(
            (json.decoder.JSONDecodeError,
             jsonschema.ValidationError)
        ),
    )
    def enforced_function_call(self,
                               messages: Messages,
                               function: Function,
                               *,
                               max_tokens: Optional[int] = None,
                               model_params: Optional[dict] = None,) -> dict:
        """
        This function enforces the model to respond with a specific function call.

        To read more about this feature, see: https://platform.openai.com/docs/guides/gpt/function-calling

        Note: this function is wrapped in a retry decorator to handle transient errors.

        Args:
            messages: Messages (chat history) to send to the model.
            function: Function to call. See canopy.llm.models.Function for more details.
            max_tokens: Maximum number of tokens to generate. Defaults to None (generates until stop sequence or until hitting max context size).
            model_params: Model parameters to use for this request. Defaults to None (uses the default model parameters).
                          Overrides the default model parameters if set on initialization.
                          For example, you can pass: {"temperature": 0.9, "top_p": 1.0} to override the default temperature and top_p.
                          see: https://platform.openai.com/docs/api-reference/chat/create

        Returns:
            dict: Function call arguments as a dictionary.

        Usage:
            >>> from canopy.llm import OpenAILLM
            >>> from canopy.llm.models import Function, FunctionParameters, FunctionArrayProperty
            >>> from canopy.models.data_models import UserMessage
            >>> llm = OpenAILLM()
            >>> messages = [UserMessage(content="I was wondering what is the capital of France?")]
            >>> function = Function(
            ...     name="query_knowledgebase",
            ...     description="Query search engine for relevant information",
            ...     parameters=FunctionParameters(
            ...         required_properties=[
            ...             FunctionArrayProperty(
            ...                 name="queries",
            ...                 items_type="string",
            ...                 description='List of queries to send to the search engine.',
            ...             ),
            ...         ]
            ...     )
            ... )
            >>> llm.enforced_function_call(messages, function)
            {'queries': ['capital of France']}
        """  # noqa: E501

        model_params_dict: Dict[str, Any] = deepcopy(self.default_model_params)
        model_params_dict.update(
            model_params or {}
        )

        function_dict = cast(ChatCompletionToolParam,
                             {"type": "function", "function": function.dict()})

        try:
            chat_completion = self._client.chat.completions.create(
                messages=[m.dict() for m in messages],
                model=self.model_name,
                tools=[function_dict],
                tool_choice={"type": "function",
                             "function": {"name": function.name}},
                max_tokens=max_tokens,
                **model_params_dict
            )
        except openai.OpenAIError as e:
            provider_name = self.__class__.__name__.replace("LLM", "")
            raise RuntimeError(
                f"Failed to use {provider_name}'s {self.model_name} model for "
                f"chat completion with enforced function calling.\n"
                f"Error: {_format_openai_error(e)}"
            )

        result = chat_completion.choices[0].message.tool_calls[0].function.arguments
        arguments = json.loads(result)

        jsonschema.validate(instance=arguments, schema=function.parameters.dict())
        return arguments

    async def achat_completion(self,
                               messages: Messages, *, stream: bool = False,
                               max_generated_tokens: Optional[int] = None,
                               model_params: Optional[dict] = None,
                               ) -> Union[ChatResponse,
                                          Iterable[StreamingChatChunk]]:
        raise NotImplementedError()

    async def agenerate_queries(self,
                                messages: Messages,
                                *,
                                max_generated_tokens: Optional[int] = None,
                                model_params: Optional[dict] = None,
                                ) -> List[Query]:
        raise NotImplementedError()



from typing import List

from .langchain_text_splitter import Language, RecursiveCharacterTextSplitter
from .recursive_character import RecursiveCharacterChunker
from canopy.knowledge_base.models import KBDocChunk
from canopy.models.data_models import Document


class MarkdownChunker(RecursiveCharacterChunker):
    """
    MarkdownChunker is a subclass of RecursiveCharacterChunker that is configured
    to chunk markdown documents. It uses RecursiveCharacterTextSplitter to split
    the text of the document into chunks, by providing the separators for markdown documents
    (also from LangChainTextSplitter, with modifications)
    """  # noqa: E501

    def __init__(self,
                 chunk_size: int = 256,
                 chunk_overlap: int = 0,
                 keep_separator: bool = True
                 ):
        """
        Iniitalizes RecursiveCharacterChunker with the separators for markdown documents.

        Args:
            chunk_size: size of the chunks. Defaults to 256 tokens.
            chunk_overlap: overlap between chunks. Defaults to 0.
            keep_separator: whether to keep the separator in the chunk. Defaults to True.

        """  # noqa: E501
        separators = RecursiveCharacterTextSplitter.get_separators_for_language(
            Language.MARKDOWN
        )
        super().__init__(chunk_size=chunk_size,
                         chunk_overlap=chunk_overlap,
                         separators=separators,
                         keep_separator=keep_separator)

    async def achunk_single_document(self,
                                     document: Document) -> List[KBDocChunk]:
        raise NotImplementedError()



from abc import ABC, abstractmethod
from typing import Tuple

from canopy.tokenizer import Tokenizer
from canopy.models.data_models import Messages


class HistoryPruner(ABC):

    def __init__(self,
                 min_history_messages: int):
        self._tokenizer = Tokenizer()
        self._min_history_messages = min_history_messages

    @abstractmethod
    def build(self,
              full_history: Messages,
              max_tokens: int) -> Tuple[Messages, int]:
        raise NotImplementedError

    async def abuild(self,
                     full_history: Messages,
                     max_tokens: int) -> Tuple[Messages, int]:
        raise NotImplementedError()



from typing import List
from .base import BaseTokenizer
from ..models.data_models import Messages
import os
from transformers import LlamaTokenizerFast as HfTokenizer


class LlamaTokenizer(BaseTokenizer):
    """
    Tokenizer for Llama models, based on the tokenizers library.

    Usage:
    Initialize the singleton tokenizer with the LlamaTokenizer class:
    >>> from canopy.tokenizer import Tokenizer
    >>> Tokenizer.initialize(tokenizer_class=LlamaTokenizer,
                             model_name="hf-internal-testing/llama-tokenizer")

    You can then use the tokenizer instance from anywhere in the code:
    >>> from canopy.tokenizer import Tokenizer
    >>> tokenizer = Tokenizer()
    >>> tokenizer.tokenize("Hello World!")
    ['▁Hello', '▁World', '!']
    """  # noqa: E501

    MESSAGE_TOKENS_OVERHEAD = 3
    FIXED_PREFIX_TOKENS = 3

    def __init__(
        self,
        model_name: str = "hf-internal-testing/llama-tokenizer",
        hf_token: str = "",
    ):
        """
        Initialize the tokenizer.

        Args:
            model_name: The name of the model to use. Defaults to "openlm-research/open_llama_7b_v2".
            hf_token: Huggingface token
        """  # noqa: E501
        hf_token = hf_token or os.environ.get("HUGGINGFACE_TOKEN", "")
        # Add legacy=True to avoid extra printings
        self._encoder = HfTokenizer.from_pretrained(
            model_name, token=hf_token, legacy=True, add_bos_token=False
        )

    def tokenize(self, text: str) -> List[str]:
        """
        Tokenize a text using tiktoken.

        Args:
            text: The text to tokenize.

        Returns:
            The list of tokens.
        """
        return self._encoder.tokenize(text)

    def detokenize(self, tokens: List[str]) -> str:
        """
        Detokenize a list of tokens that were previously tokenized using this tokenizer.

        Args:
            tokens: The list of tokens to detokenize.

        Returns:
            The detokenized text as a string.
        """
        return self._encoder.convert_tokens_to_string(tokens)

    def token_count(self, text: str) -> int:
        """
        Count the number of tokens in a text.

        Args:
            text: The text to count the tokens of.

        Returns:
            The number of tokens in the text.
        """
        return len(self._encode(text))

    def _encode(self, text):
        # Return Encoding objects, which contains attributes ids and tokens
        return self._encoder.encode(text)

    def messages_token_count(self, messages: Messages) -> int:
        """
        Count the number of tokens in a list of messages as expected to be counted by OpenAI models.
        Account for the overhead of the messages structure.
        Taken from: https://github.com/openai/openai-cookbook/.../How_to_format_inputs_to_ChatGPT_models.ipynb

        Args:
            messages: The list of messages to count the tokens of.

        Returns:
            The number of tokens in the messages, as expected to be counted by OpenAI models.
        """  # noqa: E501
        num_tokens = 0
        for message in messages:
            num_tokens += self.MESSAGE_TOKENS_OVERHEAD
            for key, value in message.dict().items():
                num_tokens += self.token_count(value)
        num_tokens += self.FIXED_PREFIX_TOKENS
        return num_tokens



from typing import Union, Iterable, Optional, Any, List
import jsonschema
import json
import os
from tenacity import (
    retry,
    stop_after_attempt,
    retry_if_exception_type,
)
from canopy.llm import OpenAILLM
from canopy.llm.models import Function
from canopy.models.api_models import ChatResponse, StreamingChatChunk
from canopy.models.data_models import Messages, Query

ANYSCALE_BASE_URL = "https://api.endpoints.anyscale.com/v1"


class AnyscaleLLM(OpenAILLM):
    """
    Anyscale LLM wrapper built on top of the OpenAI Python client.

    Note: Anyscale requires a valid API key to use this class.
          You can set the "ANYSCALE_API_KEY" environment variable.
    """

    def __init__(
        self,
        model_name: str = "meta-llama/Llama-2-7b-chat-hf",
        *,
        base_url: Optional[str] = ANYSCALE_BASE_URL,
        api_key: Optional[str] = None,
        **kwargs: Any,
    ):
        ae_api_key = api_key or os.environ.get("ANYSCALE_API_KEY")
        if not ae_api_key:
            raise ValueError(
                "Anyscale API key is required to use Anyscale. "
                "Please provide it as an argument "
                "or set the ANYSCALE_API_KEY environment variable."
            )
        ae_base_url = base_url
        super().__init__(model_name, api_key=ae_api_key, base_url=ae_base_url, **kwargs)

    @retry(
        reraise=True,
        stop=stop_after_attempt(3),
        retry=retry_if_exception_type(
            (json.decoder.JSONDecodeError, jsonschema.ValidationError)
        ),
    )
    def enforced_function_call(
        self,
        messages: Messages,
        function: Function,
        *,
        max_tokens: Optional[int] = None,
        model_params: Optional[dict] = None,
    ) -> dict:
        raise NotImplementedError()

    async def achat_completion(
        self,
        messages: Messages,
        *,
        stream: bool = False,
        max_generated_tokens: Optional[int] = None,
        model_params: Optional[dict] = None,
    ) -> Union[ChatResponse, Iterable[StreamingChatChunk]]:
        raise NotImplementedError()

    async def agenerate_queries(
        self,
        messages: Messages,
        *,
        max_generated_tokens: Optional[int] = None,
        model_params: Optional[dict] = None,
    ) -> List[Query]:
        raise NotImplementedError()



from abc import ABC, abstractmethod
from typing import List

from canopy.models.data_models import Messages
from canopy.utils.config import ConfigurableMixin


class BaseTokenizer(ABC, ConfigurableMixin):

    @abstractmethod
    def tokenize(self, text: str) -> List[str]:
        pass

    @abstractmethod
    def detokenize(self, tokens: List[str]) -> str:
        pass

    def token_count(self, text: str) -> int:
        return len(self.tokenize(text))

    @abstractmethod
    def messages_token_count(self, messages: Messages) -> int:
        pass
